"""
  Classes for various discrete probability distributions
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import
# pylint: disable=invalid-name
# pylint: disable=abstract-method
# pylint: disable=relative-import

import numpy as np

# Local imports
from .distribution import Discrete
from ..exd import domains

# Bernoulli Distribution
class Bernoulli(Discrete):
  """ Bernoulli Distribution """

  def __init__(self, p):
    """ Constructor. """
    super(Bernoulli, self).__init__()
    self.p = float(p)
    self.dim = 1
    self.domain = domains.IntegralDomain([[0, 1]])

  def pmf(self, x):
    """ Returns pmf of distribution at x. """
    return np.asscalar(self.p*x + (1 - x)*(1 - self.p))

  def draw_samples_random(self, size=None):
    """ Returns random samples drawn from distribution. """
    return np.random.binomial(1, self.p, size)

  def get_parameters(self):
    """ Returns parameters """
    return self.p

  def logp(self, x):
    """ Returns log probability of x """
    if x < 0 or x > 1:
      return -np.inf
    return x*np.log(self.p) + (1-x)*np.log(1-self.p)

  def get_mean(self):
    """ Returns mean """
    return self.p

  def get_variance(self):
    """ Returns variance """
    return self.p*(1 - self.p)

  def __str__(self):
    """ Returns a string representation. """
    return 'Bernoulli: p=%0.3f' % (self.p)


# Binomial Distribution
class Binomial(Discrete):
  """ Binomial Distribution """

  def __init__(self, n, p):
    """ Constructor. """
    super(Binomial, self).__init__()
    self.n = n
    self.p = float(p)
    self.dim = 1
    self.domain = domains.IntegralDomain([[0, self.n]])

  def pmf(self, k):
    """ Returns pmf of distribution at x. """
    value = np.math.factorial(self.n)
    value = value/float((np.math.factorial(k)*np.math.factorial(self.n-k)))
    return np.asscalar(value*np.power(self.p, k)*np.power(1-self.p, self.n-k))

  def draw_samples_random(self, size=None):
    """ Returns random samples drawn from distribution. """
    return np.random.binomial(self.n, self.p, size)

  def get_parameters(self):
    """ Returns parameters """
    return self.n, self.p

  def logp(self, x):
    """ Returns log probability of x """
    if x < 0 or x > self.n:
      return -np.inf

    value = np.math.factorial(self.n)
    value = value/float(np.math.factorial(x)*np.math.factorial(self.n-x))
    return np.log(value) + x*np.log(self.p) + (self.n - x)*np.log(1-self.p)

  def get_mean(self):
    """ Returns mean """
    return self.n*self.p

  def get_variance(self):
    """ Returns variance """
    return self.n*self.p*(1 - self.p)

  def __str__(self):
    """ Returns a string representation. """
    return 'Binomial: n=%0.3f, p=%0.3f' % (self.n, self.p)


# Categorical Distribution
class Categorical(Discrete):
  """ Categorical Distribution """

  def __init__(self, categories, p):
    """ Constructor """
    super(Categorical, self).__init__()
    self.cat = list(categories)
    self.k = len(categories)
    self.p = np.array(p)
    self.dim = 1
    self.domain = domains.IntegralDomain([[1, self.k]])

  def pmf(self, i):
    """ Returns pmf of distribution at x. """
    if i < 0 or i >= self.k:
      return 0
    return np.asscalar(self.p[i])

  def draw_samples_random(self, size=None):
    """ Returns random samples drawn from distribution. """
    samples = np.random.multinomial(1, self.p, size)
    return np.argmax(samples, len(samples.shape) - 1)

  def get_category(self, i):
    """ Returns category value. """
    if i < 0 or i >= self.k:
      return None
    return self.cat[int(i)]

  def get_parameters(self):
    """ Returns parameters """
    return self.cat, self.p

  def logp(self, value):
    """ Returns log probability of category """
    if hasattr(value, '__len__'):
      if len(value) == 1:
        value = np.asscalar(value)
      else:
        raise ValueError('Input dimension should be 1.')

    if value < 0 or value >= self.k:
      return -np.inf
    return np.log(self.p[value])

  def get_id(self, category):
    """Returns id for the category. """
    if category is None or np.isnan(category):
      return -1
    return self.cat.index(category)

  def __str__(self):
    """ Returns a string representation. """
    return 'Categorical: k=%0.3f, p=%0.3f' % (self.k, self.p)


# Discrete Uniform Discrete
class DiscreteUniform(Discrete):
  """ Discrete Uniform Distribution """

  def __init__(self, lower, upper):
    """ Constructor. """
    super(DiscreteUniform, self).__init__()
    self.lower = np.floor(lower)
    self.upper = np.floor(upper)
    self.dim = 1
    self.domain = domains.IntegralDomain([[self.lower, self.upper]])

  def pmf(self, x):
    """ Returns pmf of distribution at x. """
    if x >= self.lower and x <= self.upper:
      return 1/(self.upper - self.lower + 1)
    return 0

  def draw_samples_random(self, size=None):
    """ Returns random samples drawn from distribution. """
    return np.random.randint(self.lower, self.upper+1, size)

  def get_parameters(self):
    """ Returns parameters """
    return self.lower, self.upper

  def logp(self, x):
    """ Returns log probability of x """
    if x < self.lower or x > self.upper:
      return -np.inf

    return -np.log(self.upper - self.lower + 1)

  def get_mean(self):
    """ Returns mean """
    return (self.lower + self.upper)/2

  def get_variance(self):
    """ Returns variance """
    return (np.power((self.upper - self.lower + 1), 2) - 1)/12

  def __str__(self):
    """ Returns a string representation. """
    return 'Discrete Uniform: lower=%0.3f, upper=%0.3f' % (self.lower, self.upper)




"""
  Implements GPs for Euclidean spaces.
  -- kandasamy@cs.cmu.edu
  -- kvysyara@andrew.cmu.edu
"""

from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=attribute-defined-outside-init
# pylint: disable=no-member

from argparse import Namespace
import numpy as np
# Local imports
from . import gp_core, mf_gp
from . import kernel as gp_kernel
from ..utils.ancillary_utils import get_list_of_floats_as_str
from ..utils.general_utils import get_sublist_from_indices, map_to_bounds
from ..utils.option_handler import get_option_specs, load_options
from ..utils.oper_utils import random_sample_from_discrete_domain
from ..utils.reporters import get_reporter

_DFLT_KERNEL_TYPE = 'se'

# Some basic parameters for Euclidean GPs.
basic_euc_gp_args = [ \
  get_option_specs('kernel_type', False, 'default',
                   'Specify type of kernel. This depends on the application.'),
  get_option_specs('use_same_bandwidth', False, False,
                   ('If true, will use same bandwidth on all dimensions. Applies only '
                    'when kernel_type is se or matern. Default=False.')), \
  ]
# Parameters for the SE kernel.
se_gp_args = [ \
  ]
# Parameters for the matern kernel
matern_gp_args = [ \
  get_option_specs('matern_nu', False, -1.0, \
    ('Specify the nu value for the matern kernel. If negative, will fit.')),
                 ]
# Parameters for the Polynomial kernel.
poly_gp_args = [ \
  get_option_specs('use_same_scalings', False, False,
                   'If true uses same scalings on all dimensions. Default is False.'),
  get_option_specs('poly_order', False, 1,
                   'Order of the polynomial to be used. Default is 1 (linear kernel).')
               ]
# Parameters for an additive kernel
add_gp_args = [ \
  get_option_specs('use_additive_gp', False, False,
                   'Whether or not to use an additive GP. '),
  get_option_specs('add_max_group_size', False, 6,
                   'The maximum number of groups in the additive grouping. '),
  get_option_specs('add_grouping_criterion', False, 'randomised_ml',
                   'Specify the grouping algorithm, should be one of {randomised_ml}'),
  get_option_specs('num_groups_per_group_size', False, -1,
                   'The number of groups to try per group size.'),
  get_option_specs('add_group_size_criterion', False, 'sampled',
                   'Specify how to pick the group size, should be one of {max, sampled}.')
              ]
# Parameters for an esp kernel
esp_gp_args = [ \
  get_option_specs('esp_order', False, -1,
                   'Order of the esp kernel. '),
  get_option_specs('esp_kernel_type', False, 'se',
                   'Specify type of kernel. This depends on the application.'),
  get_option_specs('esp_matern_nu', False, -1.0, \
                   ('Specify the nu value for matern kernel. If negative, will fit.')),
              ]
# All parameters for a basic GP
euclidean_gp_args = gp_core.mandatory_gp_args + basic_euc_gp_args + se_gp_args + \
                    matern_gp_args + poly_gp_args + add_gp_args + esp_gp_args

# Hyper-parameters for Euclidean Multi-fidelity GPs
basic_mf_euc_gp_args = [ \
  # Fidelity kernel ------------------------------------------------------------
    get_option_specs('fidel_kernel_type', False, 'se', \
      'Type of kernel for the fidelity space. Should be se, matern, poly or expdecay'),
    # Secondary parameters for the fidelity kernel
    get_option_specs('fidel_matern_nu', False, 2.5, \
      ('Specify the nu value for the matern kernel. If negative, will fit.')),
    get_option_specs('fidel_use_same_bandwidth', False, False, \
      ('If true, will use same bandwidth on all fidelity dimensions. Applies only when ' \
      'fidel_kernel_type is se or matern. Default=False.')),
    get_option_specs('fidel_use_same_scalings', False, False, \
      ('If true, will use same scaling on all fidelity dimensions. Applies only when ' \
      'fidel_kernel_type is poly. Default=False.')),
    get_option_specs('fidel_poly_order', False, -1, \
      ('Order of the polynomial for fidelity kernel. Default = -1 (means will tune)')),
    # Domain kernel --------------------------------------------------------------
    get_option_specs('domain_kernel_type', False, 'se',
                     'Type of kernel for the domain space. Should be se, matern or poly'),
    # Secondary parameters for the domain kernel
    get_option_specs('domain_matern_nu', False, 2.5, \
      ('Specify the nu value for the matern kernel. If negative, will fit.')),
    get_option_specs('domain_use_same_bandwidth', False, False, \
      ('If true, will use same bandwidth on all domain dimensions. Applies only when ' \
      'domain_kernel_type is se or matern. Default=False.')),
    get_option_specs('domain_use_same_scalings', False, False, \
      ('If true, will use same scaling on all domainity dimensions. Applies only when ' \
      'domain_kernel_type is poly. Default=False.')),
    get_option_specs('domain_poly_order', False, -1, \
      ('Order of the polynomial for domainity kernel. Default = -1 (means will fit)')),
    # Additive models for the domain kernel
    get_option_specs('domain_use_additive_gp', False, False,
                     'Whether or not to use an additive GP. '),
    get_option_specs('domain_add_max_group_size', False, 6,
                     'The maximum number of groups in the additive grouping. '),
    get_option_specs('domain_add_grouping_criterion', False, 'randomised_ml',
                     'Specify the grouping algorithm, should be one of {randomised_ml}'),
    get_option_specs('domain_num_groups_per_group_size', False, -1,
                     'The number of groups to try per group size.'),
    get_option_specs('domain_add_group_size_criterion', False, 'sampled', \
      'Specify how to pick the group size, should be one of {max, sampled}.'), \
    get_option_specs('domain_esp_order', False, -1,
                     'Order of the esp kernel. '),
    get_option_specs('domain_esp_kernel_type', False, 'se',
                     'Specify type of kernel. This depends on the application.'),
    get_option_specs('domain_esp_matern_nu', False, -1.0, \
                     ('Specify the nu value for matern kernel. If negative, will fit.')),
    get_option_specs('fidel_esp_order', False, -1,
                     'Order of the esp kernel. '),
    get_option_specs('fidel_esp_kernel_type', False, 'se',
                     'Specify type of kernel. This depends on the application.'),
    get_option_specs('fidel_esp_matern_nu', False, -1.0, \
                     ('Specify the nu value for matern kernel. If negative, will fit.')),\
  ]
# Define this which includes mandatory_gp_args and basic_gp_args
euclidean_mf_gp_args = gp_core.mandatory_gp_args + basic_mf_euc_gp_args

# Part I: EuclideanGP and EuclideanGPFitter
# ======================================================================================
class EuclideanGP(gp_core.GP):
  """ euclidean GP factory """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, X, Y, kernel, mean_func, noise_var,
               kernel_hyperparams=None, build_posterior=True, reporter=None):
    """
    X, Y: data
    kern: could be an object or one of the following strings, 'se', 'poly', 'matern'
    kernel_hyperparams: dictionary specifying the hyper-parameters for the kernel.
          'se' : 'dim', 'scale' (optional), 'dim_bandwidths' (optional)
          'poly' : 'dim', 'order', 'scale', 'dim_scalings' (optional)
          'matern' : 'dim', 'nu' (optional), 'scale' (optional),
                     'dim_bandwidths' (optional)
    """
    if isinstance(kernel, str):
      kernel = self._get_kernel_from_type(kernel, kernel_hyperparams)
    super(EuclideanGP, self).__init__(X, Y, kernel, mean_func, noise_var,
                                      build_posterior, reporter)

  @classmethod
  def _get_kernel_from_type(cls, kernel_type, kernel_hyperparams):
    """ Get different euclidean kernels based on kernel_type"""
    if kernel_type in ['se']:
      return gp_kernel.SEKernel(kernel_hyperparams['dim'], kernel_hyperparams['scale'],
                                kernel_hyperparams['dim_bandwidths'])
    elif kernel_type in ['poly']:
      return gp_kernel.PolyKernel(kernel_hyperparams['dim'], kernel_hyperparams['order'],
                                  kernel_hyperparams['scale'],
                                  kernel_hyperparams['dim_scalings'])
    elif kernel_type in ['matern']:
      return gp_kernel.MaternKernel(kernel_hyperparams['dim'],
                                    kernel_hyperparams['nu'], kernel_hyperparams['scale'],
                                    kernel_hyperparams['dim_bandwidths'])
    elif kernel_type in ['esp']:
      return gp_kernel.ESPKernelSE(kernel_hyperparams['dim'], kernel_hyperparams['scale'],
                                   kernel_hyperparams['order'],
                                   kernel_hyperparams['dim_bandwidths'])
    else:
      raise ValueError('Cannot construct kernel from kernel_type %s.' % (kernel_type))

  def _child_str(self):
    """ String representation for child GP. """
    ke_str = self._get_kernel_str(self.kernel)
    dim = 0 if len(self.X) == 0 else len(self.X[0])
    mean_str = 'mu(0)=%0.3f'%(self.mean_func([np.zeros(dim,)])[0])
    ret = 'scale: %0.3f, %s, %s' % (self.kernel.hyperparams['scale'], ke_str, mean_str)
    return ret

  @classmethod
  def _get_kernel_str(cls, kern):
    """ Gets a string format of the kernel depending on whether it is SE/Poly."""
    if isinstance(kern, gp_kernel.AdditiveKernel):
      return str(kern)
    if isinstance(kern, gp_kernel.SEKernel) or isinstance(kern, gp_kernel.MaternKernel):
      hp_name = 'dim_bandwidths'
      kern_name = 'se' if isinstance(kern, gp_kernel.SEKernel) else \
        'matern(%0.1f)' % (kern.hyperparams['nu'])
    elif isinstance(kern, gp_kernel.PolyKernel):
      hp_name = 'dim_scalings'
      kern_name = 'poly'
    else:  # Return an empty string.
      return ''
    if kern.dim > 6:
      ret = '%0.4f(avg)' % (kern.hyperparams[hp_name].mean())
    else:
      ret = get_list_of_floats_as_str(kern.hyperparams[hp_name])
    ret = kern_name + '-' + ret
    return ret


class EuclideanGPFitter(gp_core.GPFitter):
  """ Fits a GP by tuning the kernel hyper-params. """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, X, Y, options=None, reporter=None):
    """ Constructor. """
    self.dim = len(X[0])
    reporter = get_reporter(reporter)
    if options is None:
      options = load_options(euclidean_gp_args, 'EuclideanGPFitter', reporter=reporter)
    super(EuclideanGPFitter, self).__init__(X, Y, options, reporter)

  def _child_set_up(self):
    """ Sets parameters for GPFitter. """
    # IMPORTANT: Keep this order when tuning for the hyper-parameters.
    # Continuous: Mean value, GP noise, kernel scale, then the remaining Kernel params.
    # Discrete: add_group_size, then the remaining kernel params.
    # Check args - so that we don't have to keep doing this all the time
    if self.options.kernel_type not in ['se', 'matern', 'poly', 'esp', 'default']:
      raise ValueError('Unknown kernel_type. Should be either se, matern or poly.')
    if self.options.noise_var_type not in ['tune', 'label', 'value']:
      raise ValueError('Unknown noise_var_type. Should be either tune, label or value.')
    if self.options.mean_func_type not in ['mean', 'median', 'const', 'zero', 'tune']:
      raise ValueError('Unknown mean_func_type. Should be mean/median/const/zero/tune.')
    # Set kernel type
    if self.options.kernel_type == 'default':
      self.kernel_type = _DFLT_KERNEL_TYPE
    else:
      self.kernel_type = self.options.kernel_type
    # 1 & 2: mean value and noise variance - done in parent class.
    # 3. Kernel parameters
    if self.kernel_type == 'se':
      self._se_kernel_set_up()
    elif self.kernel_type == 'matern':
      self._matern_kernel_set_up()
    elif self.kernel_type in ['poly']:
      self._poly_kernel_set_up()
    elif self.kernel_type == 'esp':
      self._esp_kernel_set_up()
    # 4. Additive grouping
    if self.options.use_additive_gp and self.kernel_type != 'esp':
      self.add_group_size_idx_in_dscr_hp_vals = len(self.dscr_hp_vals)
      self.add_max_group_size = min(self.options.add_max_group_size, self.dim)
      self.dscr_hp_vals.append([x+1 for x in range(self.add_max_group_size)])
      self.param_order.append(["additive_grp", "dscr"])
    elif self.kernel_type == 'esp' and self.options.esp_order == -1:
      self.dscr_hp_vals.append(list(range(1, max(self.dim, self.options.esp_order) + 1)))
      self.param_order.append(["esp_order", "dscr"])

  def _se_kernel_set_up(self):
    """ Set up for the SE kernel. """
    # Scale
    self.scale_log_bounds = [np.log(0.1 * self.Y_var), np.log(10 * self.Y_var)]
    self.param_order.append(["scale", "cts"])
    # Bandwidths
    X_std_norm = np.linalg.norm(self.X, 'fro') + 1e-4
    single_bandwidth_log_bounds = [np.log(0.01 * X_std_norm), np.log(10 * X_std_norm)]
    if self.options.use_same_bandwidth:
      self.bandwidth_log_bounds = [single_bandwidth_log_bounds]
      self.param_order.append(["same_dim_bandwidths", "cts"])
    else:
      self.bandwidth_log_bounds = [single_bandwidth_log_bounds] * self.dim
      for _ in range(self.dim):
        self.param_order.append(["dim_bandwidths", "cts"])
    self.cts_hp_bounds += [self.scale_log_bounds] + self.bandwidth_log_bounds

  def _matern_kernel_set_up(self):
    """ Set up for the Matern kernel. """
    # Set up scale and bandwidth - which is identical to the SE kernel.
    self._se_kernel_set_up()
    # Set up optimisation values for the nu parameter.
    if self.options.matern_nu < 0:
      self.dscr_hp_vals.append([0.5, 1.5, 2.5])
      self.param_order.append(["nu", "dscr"])

  def _poly_kernel_set_up(self):
    """ Set up for the Poly kernel. """
    raise NotImplementedError('Not implemented Poly kernel yet.')

  def _esp_kernel_set_up(self):
    """ Set up for the ESP kernel. """
    if self.options.esp_kernel_type not in ['se', 'matern']:
      raise NotImplementedError('Not implemented yet.')
    # Scale
    self.scale_log_bounds = [np.log(0.1 * self.Y_var), np.log(10 * self.Y_var)]
    self.param_order.append(["scale", "cts"])
    # Bandwidths
    X_std_norm = np.linalg.norm(self.X, 'fro') + 1e-4
    single_bandwidth_log_bounds = [np.log(0.01 * X_std_norm), np.log(10 * X_std_norm)]
    self.bandwidth_log_bounds = [single_bandwidth_log_bounds] * self.dim
    for _ in range(self.dim):
      self.param_order.append(["dim_bandwidths", "cts"])
    self.cts_hp_bounds += [self.scale_log_bounds] + self.bandwidth_log_bounds
    if self.options.esp_kernel_type == 'matern' and self.options.esp_matern_nu < 0:
      self.dscr_hp_vals.append([0.5, 1.5, 2.5])
      self.param_order.append(["nu", "dscr"])

  def _prep_init_kernel_hyperparams(self, kernel_type):
    """ Wrapper to pack the kernel hyper-parameters into a dictionary. """
    return prep_euclidean_integral_kernel_hyperparams(kernel_type, self.options, self.dim)

  def _optimise_cts_hps_for_given_dscr_hps(self, given_dscr_hps):
    """ Optimises the continuous hyper-parameters for the given discrete hyper-params.
    """
    if not self.options.use_additive_gp:
      return super(EuclideanGPFitter, self)._optimise_cts_hps_for_given_dscr_hps( \
                                                                           given_dscr_hps)
    else:
      return optimise_cts_hps_for_given_dscr_hps_in_add_model(given_dscr_hps, \
        self.options.num_groups_per_group_size, self.dim, self.hp_tune_max_evals, \
        self.cts_hp_optimise, self._tuning_objective)

  def _sample_cts_dscr_hps_for_rand_exp_sampling(self):
    """ Samples continous and discrete hyper-parameters for rand_exp_sampling. """
    if not self.options.use_additive_gp:
      return super(EuclideanGPFitter, self)._sample_cts_dscr_hps_for_rand_exp_sampling()
    else:
      return sample_cts_dscr_hps_for_rand_exp_sampling_in_add_model( \
        self.hp_tune_max_evals, self.cts_hp_bounds, self.dim, self.dscr_hp_vals, \
        self.add_group_size_idx_in_dscr_hp_vals, self._tuning_objective)

  def _child_build_gp(self, mean_func, noise_var, gp_cts_hps, gp_dscr_hps,
                      other_gp_params=None, *args, **kwargs):
    """ Builds the GP. """
    # Domain kernel --------------------------------------
    kernel_hyperparams = self._prep_init_kernel_hyperparams(self.kernel_type)
    add_gp_groupings = None
    if self.options.use_additive_gp:
      gp_dscr_hps = gp_dscr_hps[:-1] # The first element is the group size
      add_gp_groupings = other_gp_params.add_gp_groupings
    kernel, gp_cts_hps, gp_dscr_hps = \
      get_euclidean_integral_gp_kernel(self.kernel_type, kernel_hyperparams, gp_cts_hps,
                                       gp_dscr_hps, self.options.use_same_bandwidth,
                                       add_gp_groupings, self.options.esp_kernel_type)
    ret_gp = EuclideanGP(self.X, self.Y, kernel, mean_func, noise_var, *args, **kwargs)
    return ret_gp, gp_cts_hps, gp_dscr_hps
  # EuclideanGPFitter ends here -------------------------------------------------------


# Part II: EuclideanMFGP and EuclideanMFGPFitter
# ======================================================================================
# MFGP and Fitter in Euclidean spaces: An instantiation of MFGP when both fidel_space and
# domain are euclidean.
class EuclideanMFGP(mf_gp.MFGP):
  """ An MFGP for Euclidean spaces. """

  def __init__(self, ZZ, XX, YY, mf_kernel,
               kernel_scale, fidel_kernel, domain_kernel,
               mean_func, noise_var, *args, **kwargs):
    """ Constructor. ZZ, XX, YY are the fidelity points, domain points and labels
        respectively.
    """
    if len(ZZ) != 0:
      self.fidel_dim = len(ZZ[0])
      self.domain_dim = len(XX[0])
    if fidel_kernel is not None and domain_kernel is not None:
      self.fidel_kernel = fidel_kernel
      self.domain_kernel = domain_kernel
      self.fidel_dim = fidel_kernel.dim
      self.domain_dim = domain_kernel.dim
    elif 'fidel_dim' in kwargs and 'domain_dim' in kwargs:
      self.fidel_dim = kwargs['fidel_dim']
      self.domain_dim = kwargs['domain_dim']
    else:
      raise Exception('Specify fidel_dim and domain_dim.')
    self.fidel_coords = list(range(self.fidel_dim))
    self.domain_coords = list(range(self.fidel_dim, self.fidel_dim + self.domain_dim))
    if mf_kernel is None:
      mf_kernel = gp_kernel.CoordinateProductKernel(self.fidel_dim + self.domain_dim, \
                                         kernel_scale, [fidel_kernel, domain_kernel], \
                                         [self.fidel_coords, self.domain_coords],)
      # Otherwise, we assume mf_kernel is already an appropriate kernel
    super(EuclideanMFGP, self).__init__(ZZ, XX, YY, mf_kernel, mean_func, noise_var,
                                        *args, **kwargs)

  def _test_fidel_domain_dims(self, test_fidel_dim, test_domain_dim):
    """ Tests if test_fidel_dim and test_domain_dim are equal to self.fidel_dim and
        self.domain_dim respectively and if not raises an error.
        Mostly for internal use. """
    if test_fidel_dim != self.fidel_dim or test_domain_dim != self.domain_dim:
      raise ValueError('ZZ, XX dimensions should be (%d, %d). Given (%d, %d)'%( \
                       self.fidel_dim, self.domain_dim, test_fidel_dim, test_domain_dim))

  def get_ZX_from_ZZ_XX(self, ZZ, XX):
    """ Gets the coordinates in the joint space from the individual fidelity and
        domain spaces. """
    ordering = np.argsort(self.fidel_coords + self.domain_coords)
    if hasattr(ZZ, '__iter__') and len(ZZ) == 0:
      return []
    elif hasattr(ZZ[0], '__iter__'):
      # A list of points
      self._test_fidel_domain_dims(len(ZZ[0]), len(XX[0]))
      ZX_unordered = np.concatenate((np.array(ZZ), np.array(XX)), axis=1)
      ZX = ZX_unordered[:, ordering]
      return list(ZX)
    else:
      # A single new point
      self._test_fidel_domain_dims(len(ZZ), len(XX))
      zx_unordered = np.concatenate((ZZ, XX))
      return zx_unordered[ordering]

  def get_domain_pts(self, data_idxs=None):
    """ Returns only the domain points. """
    data_idxs = data_idxs if data_idxs is not None else range(self.num_tr_data)
    return [self.XX[i] for i in data_idxs]

  def get_fidel_pts(self, data_idxs=None):
    """ Returns only the fidelity points. """
    data_idxs = data_idxs if data_idxs is not None else range(self.num_tr_data)
    return [self.ZZ[i] for i in data_idxs]


# A GPFitter for EuclideanMFGP objects. For now, this only considers product kernels
# between the fidel_space and domain.
class EuclideanMFGPFitter(mf_gp.MFGPFitter):
  """ A fitter for GPs in multi-fidelity optimisation. """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, ZZ, XX, YY, options=None, reporter=None):
    """ Constructor. options should either be a Namespace, a list or None. """
    reporter = get_reporter(reporter)
    if options is None:
      options = load_options(euclidean_mf_gp_args, 'MF-GP', reporter)
    self.fidel_dim = len(ZZ[0])
    self.domain_dim = len(XX[0])
    self.input_dim = self.fidel_dim + self.domain_dim
    super(EuclideanMFGPFitter, self).__init__(ZZ, XX, YY, options, reporter)

  # Child set up Methods
  # ===================================================================================
  def _child_set_up(self):
    """ Sets parameters fro GPFitter. """
    # pylint: disable=too-many-branches
    # Check args - so that we don't have to keep doing this all the time
    if self.options.fidel_kernel_type not in ['se', 'matern', 'poly', 'expdecay']:
      raise ValueError('Unknown fidel_kernel_type. Should be in {se, matern, poly, ' +
                       'expdecay.')
    if self.options.domain_kernel_type not in ['se', 'matern', 'poly']:
      raise ValueError('Unknown domain_kernel_type. Should be either se or poly.')
    if self.options.noise_var_type not in ['tune', 'label', 'value']:
      raise ValueError('Unknown noise_var_type. Should be either tune, label or value.')
    if self.options.mean_func_type not in ['mean', 'median', 'const', 'zero',
                                           'upper_bound', 'tune']:
      raise ValueError(('Unknown mean_func_type. Should be one of ',
                        'mean/median/const/zero.'))
    # Set some parameters we will be using often.
    self.ZZ_std_norm = np.linalg.norm(self.ZZ, 'fro') + 5e-5
    self.XX_std_norm = np.linalg.norm(self.XX, 'fro') + 5e-5
    self.ZX_std_norm = np.sqrt(self.ZZ_std_norm**2 + self.XX_std_norm**2)

    # Bounds for the hyper parameters
    # -------------------------------
    # Kernel scale
    self.scale_log_bounds = [np.log(0.1 * self.Y_var), np.log(10 * self.Y_var)]
    self.cts_hp_bounds.append(self.scale_log_bounds)
    self.param_order.append(["scale", "cts"])
    # Fidelity kernel
    if self.options.fidel_kernel_type == 'se':
      self._fidel_se_kernel_setup()
    elif self.options.fidel_kernel_type == 'matern':
      self._fidel_matern_kernel_setup()
    elif self.options.fidel_kernel_type == 'poly':
      self._fidel_poly_kernel_setup()
    elif self.options.fidel_kernel_type == 'expdecay':
      self._fidel_expdecay_kernel_setup()
    elif self.options.fidel_kernel_type == 'esp':
      self._fidel_esp_kernel_setup()
    # Domain kernel
    if self.options.domain_kernel_type == 'se':
      self._domain_se_kernel_setup()
    elif self.options.domain_kernel_type == 'matern':
      self._domain_matern_kernel_setup()
    elif self.options.domain_kernel_type == 'poly':
      self._domain_poly_kernel_setup()
    elif self.options.domain_kernel_type == 'esp':
      self._domain_esp_kernel_setup()
    # Additive grouping for domain kernel (this has to come after fidelity kernel set up).
    if self.options.domain_use_additive_gp:
      self.domain_add_group_size_idx_in_dscr_hp_vals = len(self.dscr_hp_vals)
      self.domain_add_max_group_size = min(self.options.domain_add_max_group_size,
                                           self.domain_dim)
      self.dscr_hp_vals.append([x+1 for x in range(self.domain_add_max_group_size)])
      self.param_order.append(["additive_grp", "dscr"])

  # Functions to set up each fidelity kernel -------------------------------------------
  def _fidel_se_kernel_setup(self):
    """ Sets up the fidelity kernel as an SE kernel. """
    self._fidel_se_matern_kernel_setup_common()

  def _fidel_matern_kernel_setup(self):
    """ Sets up the fidelity kernel as a Matern kernel. """
    self._fidel_se_matern_kernel_setup_common()
    # Set optimisation values for the nu parameter
    if self.options.fidel_matern_nu < 0:
      self.dscr_hp_vals.append([0.5, 1.5, 2.5])
      self.param_order.append(["nu", "dscr"])

  def _fidel_se_matern_kernel_setup_common(self):
    """ Common operators for setting up as a SE or Matern kernel. """
    if (hasattr(self.options, 'fidel_bandwidth_log_bounds') and
        self.options.fidel_bandwidth_log_bounds is not None):
      self.fidel_bandwidth_log_bounds = self.options.fidel_bandwidth_log_bounds
    else:
      self.fidel_bandwidth_log_bounds = self._get_bandwidth_log_bounds( \
        self.fidel_dim, self.ZX_std_norm, self.options.fidel_use_same_bandwidth)
    self.cts_hp_bounds.extend(self.fidel_bandwidth_log_bounds)
    if self.options.fidel_use_same_bandwidth:
      self.param_order.append(["same_dim_bandwidths", "cts"])
    else:
      for _ in range(self.fidel_dim):
        self.param_order.append(["dim_bandwidths", "cts"])

  def _fidel_poly_kernel_setup(self):
    """ Sets up the fidelity kernel as a Poly kernel. """
    self.fidel_scaling_log_bounds = self._get_poly_kernel_bounds(self.ZZ, self.XX, \
                                        self.options.fidel_use_same_scalings)
    self.cts_hp_bounds.extend(self.fidel_scaling_log_bounds)

  def _fidel_expdecay_kernel_setup(self):
    """ Sets up the fidelity kernel as an exponential decay kernel. """
    # offset
    if (hasattr(self.options, 'fidel_expdecay_offset_log_bounds') and
        self.options.fidel_expdecay_offset_log_bounds is not None):
      self.fidel_expdecay_offset_log_bounds = \
        self.options.fidel_expdecay_offset_log_bounds
    else:
      scale_range = self.Y_var / np.sqrt(self.num_tr_data)
      self.fidel_expdecay_offset_log_bounds = \
        [np.log(0.1 * scale_range), np.log(10 * scale_range)]
    # power log bounds
    if (hasattr(self.options, 'fidel_expdecay_power_log_bounds') and
        self.options.fidel_expdecay_power_log_bounds is not None):
      self.fidel_expdecay_power_log_bounds = \
        self.options.fidel_expdecay_power_log_bounds
    else:
      self.fidel_expdecay_power_log_bounds = \
        [[np.log(1e-1), np.log(50)]] * self.fidel_dim
    self.cts_hp_bounds.append(self.fidel_expdecay_offset_log_bounds)
    self.cts_hp_bounds.extend(self.fidel_expdecay_power_log_bounds)

  def _fidel_esp_kernel_setup(self):
    """ Sets up the fidelity kernel as ESP kernel. """
    if (hasattr(self.options, 'fidel_bandwidth_log_bounds') and
        self.options.fidel_bandwidth_log_bounds is not None):
      self.fidel_bandwidth_log_bounds = self.options.fidel_bandwidth_log_bounds
    else:
      self.fidel_bandwidth_log_bounds = self._get_bandwidth_log_bounds( \
                                        self.fidel_dim, self.ZX_std_norm, False)
    self.cts_hp_bounds.extend(self.fidel_bandwidth_log_bounds)
    for _ in range(self.fidel_dim):
      self.param_order.append(["dim_bandwidths", "cts"])
    if self.options.fidel_esp_kernel_type == 'matern' and \
       self.options.fidel_esp_matern_nu < 0:
      self.dscr_hp_vals.append([0.5, 1.5, 2.5])
      self.param_order.append(["nu", "dscr"])

  # Functions to set up each domain kernel -------------------------------------------
  def _domain_se_kernel_setup(self):
    """ Sets up the domainity kernel as an SE kernel. """
    self._domain_se_matern_kernel_setup_common()

  def _domain_matern_kernel_setup(self):
    """ Sets up the domainity kernel as a Matern kernel. """
    self._domain_se_matern_kernel_setup_common()
    # Set optimisation values for the nu parameter
    if self.options.domain_matern_nu < 0:
      self.dscr_hp_vals.append([0.5, 1.5, 2.5])
      self.param_order.append(["nu", "dscr"])

  def _domain_se_matern_kernel_setup_common(self):
    """ Sets up the domain kernel as a SE kernel. """
    if (hasattr(self.options, 'domain_bandwidth_log_bounds') and
        self.options.domain_bandwidth_log_bounds is not None):
      self.domain_bandwidth_log_bounds = self.options.domain_bandwidth_log_bounds
    else:
      self.domain_bandwidth_log_bounds = self._get_bandwidth_log_bounds( \
                                         self.domain_dim, self.ZX_std_norm, False)
    self.cts_hp_bounds.extend(self.domain_bandwidth_log_bounds)
    if self.options.domain_use_same_bandwidth:
      self.param_order.append(["same_dim_bandwidths", "cts"])
    else:
      for _ in range(self.domain_dim):
        self.param_order.append(["dim_bandwidths", "cts"])

  def _domain_poly_kernel_setup(self):
    """ Sets up the domain kernel as a Poly kernel. """
    self.domain_scaling_log_bounds = self._get_poly_kernel_bounds(self.ZZ, self.XX, \
                                        self.options.domain_use_same_scalings)
    self.cts_hp_bounds.extend(self.domain_scaling_log_bounds)

  def _domain_esp_kernel_setup(self):
    """ Sets up the domain kernel as ESP kernel. """
    if (hasattr(self.options, 'domain_bandwidth_log_bounds') and
        self.options.domain_bandwidth_log_bounds is not None):
      self.domain_bandwidth_log_bounds = self.options.domain_bandwidth_log_bounds
    else:
      self.domain_bandwidth_log_bounds = self._get_bandwidth_log_bounds( \
         self.domain_dim, self.ZX_std_norm, self.options.domain_use_same_bandwidth)
    self.cts_hp_bounds.extend(self.domain_bandwidth_log_bounds)
    for _ in range(self.domain_dim):
      self.param_order.append(["dim_bandwidths", "cts"])
    if self.options.domain_esp_kernel_type == 'matern' and \
       self.options.domain_esp_matern_nu < 0:
      self.dscr_hp_vals.append([0.5, 1.5, 2.5])
      self.param_order.append(["nu", "dscr"])

  @classmethod
  def _get_bandwidth_log_bounds(cls, dim, single_bw_bounds, use_same_bandwidth):
    """ Gets bandwidths for the SE kernel. """
    if isinstance(single_bw_bounds, float) or isinstance(single_bw_bounds, int):
      single_bw_bounds = [0.01*single_bw_bounds, 10*single_bw_bounds]
    single_bandwidth_log_bounds = [np.log(x) for x in single_bw_bounds]
    bandwidth_log_bounds = ([single_bandwidth_log_bounds] if use_same_bandwidth
                            else [single_bandwidth_log_bounds] * dim)
    return bandwidth_log_bounds

  def _get_poly_kernel_bounds(self, ZZ, XX, use_same_scalings):
    """ Gets bandwidths for the Polynomial kerne. """
    raise NotImplementedError('Yet to implement polynomial kernel.')
  # _child_set_up methods end here -----------------------------------------------------

  # fit_gp Methods
  # ====================================================================================
  def _optimise_cts_hps_for_given_dscr_hps(self, given_dscr_hps):
    """ Optimises the continuous hyper-parameters for the given discrete hyper-params.
        Overrides the methods from GPFiter
    """
    if not self.options.domain_use_additive_gp:
      return super(EuclideanMFGPFitter, self)._optimise_cts_hps_for_given_dscr_hps( \
                                                                           given_dscr_hps)
    else:
      return optimise_cts_hps_for_given_dscr_hps_in_add_model(given_dscr_hps, \
        self.options.domain_num_groups_per_group_size, self.domain_dim, \
        self.hp_tune_max_evals, self.cts_hp_optimise, self._tuning_objective)

  def _sample_cts_dscr_hps_for_rand_exp_sampling(self):
    """ Samples continous and discrete hyper-parameters for rand_exp_sampling. """
    if not self.options.domain_use_additive_gp:
      return super(EuclideanMFGPFitter, self)._sample_cts_dscr_hps_for_rand_exp_sampling()
    else:
      return sample_cts_dscr_hps_for_rand_exp_sampling_in_add_model( \
        self.hp_tune_max_evals, self.cts_hp_bounds, self.domain_dim, self.dscr_hp_vals, \
        self.domain_add_group_size_idx_in_dscr_hp_vals, self._tuning_objective)

  # build_gp Methods
  # ====================================================================================
  @classmethod
  def _prep_init_fidel_domain_kernel_hyperparams(cls, kernel_type, dim, matern_nu,
                                                 poly_order, esp_order, esp_matern_nu):
    """ Wrapper to pack the kernel hyper-parameters into a dictionary. """
    hyperparams = {}
    hyperparams['dim'] = dim
    if kernel_type == 'matern' and matern_nu > 0:
      hyperparams['nu'] = matern_nu
    elif kernel_type == 'poly':
      hyperparams['order'] = poly_order
    elif kernel_type == 'esp':
      if esp_order > 0:
        hyperparams['esp_order'] = esp_order
      if esp_matern_nu > 0:
        hyperparams['esp_matern_nu'] = esp_matern_nu
    return hyperparams

  def _prep_init_fidel_kernel_hyperparams(self):
    """ Wrapper to pack the fidelity kernel hyper-parameters into a dictionary. """
    options = self.options
    return self._prep_init_fidel_domain_kernel_hyperparams(options.fidel_kernel_type, \
                   self.fidel_dim, options.fidel_matern_nu, options.fidel_poly_order, \
                                options.fidel_esp_order, options.fidel_esp_matern_nu)

  def _prep_init_domain_kernel_hyperparams(self):
    """ Wrapper to pack the domain kernel hyper-parameters into a dictionary. """
    options = self.options
    return self._prep_init_fidel_domain_kernel_hyperparams(options.domain_kernel_type, \
                 self.domain_dim, options.domain_matern_nu, options.domain_poly_order, \
                               options.domain_esp_order, options.domain_esp_matern_nu)

  def _child_build_gp(self, mean_func, noise_var, gp_cts_hps, gp_dscr_hps,
                      other_gp_params=None, *args, **kwargs):
    """ Builds a Multi-fidelity GP from the hyper-parameters. """
    # IMPORTANT: The order of the code matters in this function. Do not change.
    # Kernel scale ---------------------------------------
    ke_scale = np.exp(gp_cts_hps[0])
    gp_cts_hps = gp_cts_hps[1:]
    # Fidelity kernel ------------------------------------
    fidel_kernel_hyperparams = self._prep_init_fidel_kernel_hyperparams()
    fidel_kernel, gp_cts_hps, gp_dscr_hps = \
      get_euclidean_integral_gp_kernel_with_scale(self.options.fidel_kernel_type, 1.0, \
        fidel_kernel_hyperparams, gp_cts_hps, gp_dscr_hps, \
        self.options.fidel_use_same_bandwidth, None, self.options.fidel_esp_kernel_type)
    # Domain kernel --------------------------------------
    # The code for the domain kernel should come after fidelity. Otherwise, its a bug.
    domain_kernel_hyperparams = self._prep_init_domain_kernel_hyperparams()
    if self.options.domain_use_additive_gp:
      gp_dscr_hps = gp_dscr_hps[:-1] # The first element is the group size
      add_gp_groupings = other_gp_params.add_gp_groupings
    else:
      add_gp_groupings = None
    domain_kernel, gp_cts_hps, gp_dscr_hps = \
      get_euclidean_integral_gp_kernel_with_scale(self.options.domain_kernel_type, 1.0, \
        domain_kernel_hyperparams, gp_cts_hps, gp_dscr_hps, \
        self.options.domain_use_same_bandwidth, add_gp_groupings, \
        self.options.domain_esp_kernel_type)
    # Construct and return MF GP
    ret_gp = EuclideanMFGP(self.ZZ, self.XX, self.YY, None, ke_scale, fidel_kernel,
                           domain_kernel, mean_func, noise_var, reporter=self.reporter)
    return ret_gp, gp_cts_hps, gp_dscr_hps
  # _child_build_gp methods end here ---------------------------------------------------


# Part III: Ancillary Functions
# ===============================================

# A function to optimise continous hyperparams in an additive model-----------------------
# Used by EuclideanGPFitter and EuclideanMFGPFitter.
def optimise_cts_hps_for_given_dscr_hps_in_add_model(given_dscr_hps, \
    num_groups_per_group_size, dim, hp_tune_max_evals, cts_hp_optimise, \
    tuning_objective):
  """ Optimises the continuous hyper-parameters for an additive model. """
  group_size = given_dscr_hps[-1] # The first is the max group size
  if num_groups_per_group_size < 0:
    if group_size == 1:
      num_groups_per_group_size = 1
    else:
      num_groups_per_group_size = max(5, min(2 * dim, 25))
  grp_best_hps = None
  grp_best_val = -np.inf
  grp_best_other_params = None
  # Now try out different groups picking a random grouping each time.
  for _ in range(num_groups_per_group_size):
    rand_perm = list(np.random.permutation(dim))
    groupings = [rand_perm[i:i+group_size]
                 for i in range(0, dim, group_size)]
    other_gp_params = Namespace(add_gp_groupings=groupings)
    # _tuning_objective is usually defined in gp_core.py
    cts_tuning_objective = lambda arg: tuning_objective(arg, given_dscr_hps[:],
                                                        other_gp_params=other_gp_params)
    max_evals = int(max(500, hp_tune_max_evals/num_groups_per_group_size))
    opt_cts_val, opt_cts_hps, _ = cts_hp_optimise(cts_tuning_objective, max_evals)
    if opt_cts_val > grp_best_val:
      grp_best_val = opt_cts_val
      grp_best_hps = opt_cts_hps
      grp_best_other_params = other_gp_params
  return grp_best_val, grp_best_hps, grp_best_other_params


def sample_cts_dscr_hps_for_rand_exp_sampling_in_add_model(num_evals, cts_hp_bounds, \
    dim, dscr_hp_vals, add_group_size_idx_in_dscr_hp_vals, tuning_objective):
  # IMPORTANT: We are assuming that the add_group_size is the first
  """ Samples the hyper-paramers for an additive model. """
  agsidhp = add_group_size_idx_in_dscr_hp_vals
  sample_cts_hps = []
  sample_dscr_hps = []
  sample_other_gp_params = []
  sample_obj_vals = []
  for _ in range(num_evals):
    group_size = np.random.choice(dscr_hp_vals[agsidhp])
    rand_perm = list(np.random.permutation(dim))
    groupings = [rand_perm[i:i+group_size] for i in range(0, dim, group_size)]
    curr_other_gp_params = Namespace(add_gp_groupings=groupings)
    curr_dscr_hps = random_sample_from_discrete_domain(dscr_hp_vals)
    curr_dscr_hps[agsidhp] = group_size
    curr_cts_hps = map_to_bounds(np.random.random((len(cts_hp_bounds),)), cts_hp_bounds)
    curr_obj_val = tuning_objective(curr_cts_hps, curr_dscr_hps, curr_other_gp_params)
    # Now add to the lists
    sample_cts_hps.append(curr_cts_hps)
    sample_dscr_hps.append(curr_dscr_hps)
    sample_other_gp_params.append(curr_other_gp_params)
    sample_obj_vals.append(curr_obj_val)
  sample_probs = np.exp(sample_obj_vals)
  sample_probs = sample_probs / sample_probs.sum()
  return sample_cts_hps, sample_dscr_hps, sample_other_gp_params, sample_probs


# Utilities for building Euclidean/Integral kernels
def prep_euclidean_integral_kernel_hyperparams(kernel_type, gp_fitter_options,
                                               domain_dim):
  """ Wrapper to pack the kernel hyper-parameters into a dictionary. """
  hyperparams = {}
  hyperparams['dim'] = domain_dim
  if kernel_type == 'matern' and gp_fitter_options.matern_nu > 0:
    hyperparams['nu'] = gp_fitter_options.matern_nu
  elif kernel_type == 'poly':
    hyperparams['order'] = gp_fitter_options.poly_order
  elif kernel_type == 'esp':
    if gp_fitter_options.esp_order > 0:
      hyperparams['esp_order'] = gp_fitter_options.esp_order
    if gp_fitter_options.esp_matern_nu > 0:
      hyperparams['esp_matern_nu'] = gp_fitter_options.esp_matern_nu
  return hyperparams


# A function to obtain a Euclidean GP kernel from arguments -----------------------
def get_euclidean_integral_gp_kernel(kernel_type, kernel_hyperparams, gp_cts_hps,
                                     gp_dscr_hps, use_same_bandwidth,
                                     add_gp_groupings=None, esp_kernel_type=None):
  """ A method to parse the GP kernel from hyper-parameters. Assumes the scale is
      part of the hyper-parameters as well. """
  scale = np.exp(gp_cts_hps[0]) # First the scale and then the rest of the hyper-params.
  gp_cts_hps = gp_cts_hps[1:]
  return get_euclidean_integral_gp_kernel_with_scale(kernel_type, scale, \
    kernel_hyperparams, gp_cts_hps, gp_dscr_hps, use_same_bandwidth, add_gp_groupings, \
    esp_kernel_type)

# The following function is used by EuclideanGPFitter and EuclideanMFGPFitter.
def get_euclidean_integral_gp_kernel_with_scale(kernel_type, scale, kernel_hyperparams,
                                                gp_cts_hps, gp_dscr_hps,
                                                use_same_bandwidth, add_gp_groupings=None,
                                                esp_kernel_type=None):
  """ A method to parse the GP kernel from hyper-parameters. Assumes the scale is
      not part of the hyper-parameters but given as an argument. """
  # pylint: disable=too-many-branches
  # pylint: disable=too-many-statements
  dim = kernel_hyperparams['dim']
  # For the additive ones ---------------------
  esp_order = None
  if kernel_type == 'esp':
    if 'esp_order' in kernel_hyperparams:
      esp_order = kernel_hyperparams['esp_order']
    else:
      esp_order = gp_dscr_hps[-1]
      gp_dscr_hps = gp_dscr_hps[:-1]
  is_additive = False
  if add_gp_groupings is None:
    add_gp_groupings = [list(range(dim))]
    grp_scale = scale
  elif esp_order is None:
    is_additive = True
    grp_scale = 1.0
  # Extract features and create a placeholder for the kernel constructors.
  # Some common functionality for the se, matern and poly kernels.
  if kernel_type in ['se', 'matern', 'poly']:
    if use_same_bandwidth:
      ke_dim_bandwidths = [np.exp(gp_cts_hps[0])] * dim
      gp_cts_hps = gp_cts_hps[1:]
    else:
      ke_dim_bandwidths = np.exp(gp_cts_hps[0:dim])
      gp_cts_hps = gp_cts_hps[dim:]
  elif kernel_type in ['esp']:
    if use_same_bandwidth:
      esp_dim_bandwidths = np.exp(gp_cts_hps[0:dim])
      gp_cts_hps = gp_cts_hps[dim:]
    else:
      esp_dim_bandwidths = np.exp(gp_cts_hps[0:dim*dim])
      gp_cts_hps = gp_cts_hps[dim*dim:]
  # Now iterate through the kernels for the rest of the parameters.
  if kernel_type == 'se':
    grp_kernels = [gp_kernel.SEKernel(dim=len(grp), scale=grp_scale, \
                     dim_bandwidths=get_sublist_from_indices(ke_dim_bandwidths, grp))
                   for grp in add_gp_groupings]
  elif kernel_type == 'matern':
    if 'nu' not in kernel_hyperparams or kernel_hyperparams['nu'] < 0:
      matern_nu = gp_dscr_hps[0]
      gp_dscr_hps = gp_dscr_hps[1:]
    else:
      matern_nu = kernel_hyperparams['nu']
    grp_kernels = [gp_kernel.MaternKernel(dim=len(grp), nu=matern_nu, scale=grp_scale, \
                     dim_bandwidths=get_sublist_from_indices(ke_dim_bandwidths, grp))
                   for grp in add_gp_groupings]
  elif kernel_type == 'poly':
    if 'order' not in kernel_hyperparams or kernel_hyperparams['order'] < 0:
      poly_order = gp_dscr_hps[0]
      gp_dscr_hps = gp_dscr_hps[1:]
    else:
      poly_order = kernel_hyperparams['order']
    grp_kernels = [gp_kernel.PolyKernel(dim=len(grp), order=poly_order, scale=grp_scale, \
                     dim_scalings=get_sublist_from_indices(ke_dim_bandwidths, grp))
                   for grp in add_gp_groupings]
  elif kernel_type == 'expdecay':
    exp_decay_offset = np.exp(gp_cts_hps[0])
    exp_decay_powers = np.exp(gp_cts_hps[1:dim+1])
    gp_cts_hps = gp_cts_hps[dim+1:]
    grp_kernels = [gp_kernel.ExpDecayKernel(dim=len(grp), scale=grp_scale, \
                    offset=exp_decay_offset, powers=exp_decay_powers)
                   for grp in add_gp_groupings]
  elif kernel_type == 'esp':
    if not np.isscalar(esp_order):
      esp_order = np.asscalar(esp_order)
    if esp_kernel_type == 'se':
      grp_kernels = [gp_kernel.ESPKernelSE(dim=dim, scale=scale, order=int(esp_order),
                                           dim_bandwidths=esp_dim_bandwidths)]
    elif esp_kernel_type == 'matern':
      if 'esp_matern_nu' not in kernel_hyperparams:
        matern_nu = [gp_dscr_hps[0]] * dim
        gp_dscr_hps = gp_dscr_hps[1:]
      else:
        matern_nu = [kernel_hyperparams['esp_matern_nu']] * dim
      grp_kernels = [gp_kernel.ESPKernelMatern(dim=dim, nu=matern_nu,
                                               scale=scale, order=int(esp_order),
                                               dim_bandwidths=esp_dim_bandwidths)]
  else:
    raise Exception('Unknown kernel type %s!'%(kernel_type))
  if is_additive:
    euc_kernel = gp_kernel.AdditiveKernel(scale=scale, kernel_list=grp_kernels,
                                          groupings=add_gp_groupings)
  else:
    euc_kernel = grp_kernels[0]
  return euc_kernel, gp_cts_hps, gp_dscr_hps





"""
  Unit tests for GP Bandits on Cartesian Product domains.
  -- kandasamy@cs.cmu.edu
"""
import unittest

# Local imports
from . import gp_bandit
from .unittest_cp_random_optimiser import CPOptimiserBaseTestCase
from ..utils.base_test_class import BaseTestClass, execute_tests


class CPGPBanditTestCaseDefinitions(object):
  """ Unit tests for GP Bandits on cartesian product spaces. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager, options, reporter):
    """ Instantiate the optimiser. """
    return gp_bandit.CPGPBandit(func_caller, worker_manager, options=options,
                                reporter=reporter)

  @classmethod
  def _run_optimiser(cls, prob_funcs, domain_config_file, worker_manager, max_capital,
                     mode, *args, **kwargs):
    """ Run the optimiser. """
    return gp_bandit.cp_gpb_from_raw_args(prob_funcs[0], domain_config_file,
                                          worker_manager=worker_manager,
                                          max_capital=max_capital, is_mf=False,
                                          mode=mode, *args, **kwargs)


@unittest.skip
class CPGPBanditTestCase(CPGPBanditTestCaseDefinitions,
                         CPOptimiserBaseTestCase,
                         BaseTestClass):
  """ Unit tests for GP Bandits on cartesian product spaces. """
  pass


if __name__ == '__main__':
  execute_tests()




"""
  Implements a GP for neural networks. The kernels and distances are taken from otmann.py.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=abstract-class-not-used
# pylint: disable=arguments-differ

import numpy as np
# Local
from . import otmann
from ..gp import gp_core
from ..utils.ancillary_utils import get_list_of_floats_as_str
from ..utils.reporters import get_reporter
from ..utils.option_handler import get_option_specs, load_options

_DFLT_KERNEL_TYPE = 'lpemd_sum'

nn_gp_specific_args = [
  get_option_specs('dist_type', False, 'lp-emd',
    'The type of distance. This should be lp, emd or lp-emd.'),
# Use given coeffcients by default
  get_option_specs('choose_mislabel_struct_coeffs', False, 'use_given',
    ('How to choose the mislabel and struct coefficients. Should be one of ' +
     'tune_coeffs or use_given. In the latter case, mislabel_coeffs and struct_coeffs ' +
     'should be non-empty.')),
  get_option_specs('compute_kernel_from_dists', False, True,
    'Should you compute the kernel from pairwise distances whenever possible.'),
  get_option_specs('mislabel_coeffs', False, '1.0-1.0-1.0-1.0',
    'The mislabel coefficients specified as a string. If -1, it means we will tune.'),
  get_option_specs('struct_coeffs', False, '0.1-0.25-0.61-1.5',
    'The struct coefficients specified as a string. If -1, it means we will tune.'),
  get_option_specs('lp_power', False, 1,
    'The powers to use in the LP distance for the kernel.'),
  get_option_specs('emd_power', False, 2,
    'The powers to use in the EMD distance for the kernel.'),
  get_option_specs('non_assignment_penalty', False, 1.0,
    'The non-assignment penalty.'),
  ]

nn_gp_args = gp_core.mandatory_gp_args + nn_gp_specific_args


# Main GP class for neural network architectures -----------------------------------------
class NNGP(gp_core.GP):
  """ A Gaussian process for Neural Networks. """

  def __init__(self, X, Y, kernel, mean_func, noise_var, nn_type, kernel_hyperparams=None,
               dist_type=None, tp_comp=None, list_of_dists=None, *args, **kwargs):
    # pylint: disable=too-many-arguments
    """ Constructor.
        kernel: Should be a kernel or one of 'from_distance_computer', 'lp', 'emd',
          'sum_lp-emd', 'product_lp-emd'
        kernel_hyperparams: is a python dictionary specifying the hyper-parameters for the
          kernel. Will have parameters 'beta', 'scale', 'sum_scales', 'powers'.
        list_of_dists: Is a list of distances for the training set.
        tp_comp: a transport distance computer object.
    """
    if isinstance(kernel, str):
      # then kernel is specifying the kernel type
      kernel = get_nn_kernel_from_args(kernel, nn_type, kernel_hyperparams, tp_comp,
                                       dist_type)
    self.nn_type = nn_type
    self.list_of_dists = list_of_dists
    # Call super constructor
    super(NNGP, self).__init__(X, Y, kernel, mean_func, noise_var,
                               handle_non_psd_kernels='project_first', *args, **kwargs)

  def build_posterior(self):
    """ Checks if the sizes of list of distances and the data are consistent before
        calling build_posterior from the super class.
    """
    if self.list_of_dists is not None:
      assert self.list_of_dists[0].shape == (len(self.X), len(self.Y))
    super(NNGP, self).build_posterior()

  def set_list_of_dists(self, list_of_dists):
    """ Updates the list of distances. """
    self.list_of_dists = list_of_dists

  def _get_training_kernel_matrix(self):
    """ Compute the kernel matrix from distances if they are provided. """
    if self.list_of_dists is not None:
      return self.kernel.evaluate_from_dists(self.list_of_dists)
    else:
      return self.kernel(self.X, self.X)

  def _child_str(self):
    """ Description of the child GP. """
    scales_str = self._get_scales_str()
    betas_str = 'betas=' + get_list_of_floats_as_str(self.kernel.hyperparams['betas'])
    return self.kernel.dist_type + ':: ' + scales_str + ' ' + betas_str

  def _get_scales_str(self):
    """ Returns a description of the alphas. """
    if 'alphas' in self.kernel.hyperparams.keys():
      return 'alphas=' + get_list_of_floats_as_str(self.kernel.hyperparams['alphas'])
    else:
      return 'scale=' + str(round(self.kernel.hyperparams['scale'], 4))


# GP Fitter for neural network architectures ---------------------------------------------
class NNGPFitter(gp_core.GPFitter):
  """ Fits a GP by tuning the kernel hyper-params. """
  # pylint: disable=abstract-method
  # disabling abstract-method because _set_up_post_mean_hp_tune is not implemented in
  # GPFitter.

  def __init__(self, X, Y, nn_type, tp_comp=None, list_of_dists=None,
               options=None, reporter=None, *args, **kwargs):
    """ Constructor. """
    self.nn_type = nn_type
    self.list_of_dists = list_of_dists
    self.tp_comp = tp_comp
    # get options
    reporter = get_reporter(reporter)
    if options is None:
      options = load_options(nn_gp_args, 'NNGPFitter', reporter=reporter)
    super(NNGPFitter, self).__init__(X, Y, options, reporter, *args, **kwargs)

  def _child_set_up(self):
    """ Sets up child-specific parameters for the GP Fitter. """
    # pylint: disable=too-many-branches
    # Check args
    if not self.options.dist_type in ['lp', 'emd', 'lp-emd']:
      raise ValueError('Unknown dist_type. Should be one of \'lp\', \'emd\', \'lp-emd\'.')
    if not self.options.kernel_type in ['lp', 'emd', 'lpemd_sum', 'lpemd_prod',
                                        'default']:
      raise ValueError('Unknown dist_type. Should be one of \'lp\', \'emd\', ' +
                       '\'lpemd_sum\', or \'lpemd_prod\'.')
    if not self.options.noise_var_type in ['tune', 'label', 'value']:
      raise ValueError('Unknown noise_var_type. Should be either tune, label or value.')
    if not self.options.choose_mislabel_struct_coeffs in ['use_given', 'tune_coeffs']:
      raise ValueError('Unknown value for choose_mislabel_struct_coeffs. Should be one ' +
                       'of use_given or tune_coeffs.')
    # Set kernel type
    if self.options.kernel_type == 'default':
      self.kernel_type = _DFLT_KERNEL_TYPE
    else:
      self.kernel_type = self.options.kernel_type
    # Check if dist_type and kernel_type are consistent
    if self.options.dist_type in ['lp', 'emd'] and \
       self.kernel_type != self.options.dist_type:
      raise ValueError('If dist_type is %s, then kernel_type should be %s.'%(
                       self.options.dist_type, self.options.dist_type))
    elif self.options.dist_type == 'lp-emd' and \
      self.kernel_type not in ['lpemd_prod', 'lpemd_sum']:
      raise ValueError('If dist_type is lp-emd, then kernel_type should be lpemd_sum or '
                       + 'lpemd_prod.')
    if self.options.choose_mislabel_struct_coeffs == 'use_given' and \
      (self.options.mislabel_coeffs == '' or self.options.struct_coeffs == ''):
      raise ValueError(('If choose_mislabel_struct_coeffs is use_given, then ' +
        'mislabel_coeffs and struct_coeffs cannot be empty. Given mislabel_coeffs=%s' +
        ' struct_coeffs=%s')%(self.options.mislabel_coeffs, self.options.struct_coeffs))
    self._preprocess_struct_mislabel_coeffs()
    # Check if we need to pre-compute distances
    if self.options.choose_mislabel_struct_coeffs == 'use_given' and \
       self.list_of_dists is None:
      if self.tp_comp is None:
        self.tp_comp = otmann.get_otmann_distance_computer_from_args(self.nn_type,
                                        self.options.non_assignment_penalty,
                                        self.mislabel_coeffs, self.struct_coeffs,
                                        self.options.dist_type)
      self.list_of_dists = self.tp_comp(self.X, self.X)
    # Create bounds here. Order of hyper-parameters (this is critical): scale/alphas,
    # betas, mislabel_coeff, struct_coeff
    # 1. scale/alphas -----------------------------------------
    scale_log_bounds = [np.log(0.01 * self.Y_var), np.log(100 * self.Y_var)]
    if self.kernel_type in ['lp', 'emd', 'lpemd_prod']:
      self.cts_hp_bounds.append(scale_log_bounds)
    elif self.kernel_type == 'lpemd_sum':
      self.cts_hp_bounds.extend([scale_log_bounds] * 2)
    # 2. betas ------------------------------------------------
    lp_beta_log_bounds = [[np.log(1e-9), np.log(1e-3)]] * self.num_mislabel_struct_coeffs
    emd_beta_log_bounds = [[np.log(1e-1), np.log(1e2)]] * self.num_mislabel_struct_coeffs
    if self.options.dist_type == 'lp':
      all_beta_bounds = lp_beta_log_bounds
    elif self.options.dist_type == 'emd':
      all_beta_bounds = emd_beta_log_bounds
    elif self.options.dist_type == 'lp-emd':
      all_beta_bounds = [j for i in zip(lp_beta_log_bounds, emd_beta_log_bounds)
                         for j in i]
    self.cts_hp_bounds.extend(all_beta_bounds) # extend with log_beta_bounds
    # 3 & 4. mislabel/struct_coeff ----------------------------
    if self.to_tune_mislabel_struct_coeffs:
      self.cts_hp_bounds.append([0.001, 2.0]) # mislabel coefficient (not in log space)
      self.cts_hp_bounds.append([0.001, 2.0]) # structural coefficient (not in log space)

  def _preprocess_struct_mislabel_coeffs(self):
    """ Preprocesses the structural and mislabel coefficients. """
    if self.options.choose_mislabel_struct_coeffs == 'tune_coeffs':
      # If they are integers or floats
      self.num_mislabel_struct_coeffs = 1
      self.to_tune_mislabel_struct_coeffs = True
    elif (isinstance(self.options.mislabel_coeffs, list) and
          isinstance(self.options.struct_coeffs, list)) or \
         (isinstance(self.options.mislabel_coeffs, str) and
          isinstance(self.options.struct_coeffs, str)):
      # If they are lists or strings
      if isinstance(self.options.mislabel_coeffs, str):
        # If they are strings, obtain the lists
        self.mislabel_coeffs = [float(x) for x in self.options.mislabel_coeffs.split('-')]
        self.struct_coeffs = [float(x) for x in self.options.struct_coeffs.split('-')]
      else:
        self.mislabel_coeffs = self.options.mislabel_coeffs
        self.struct_coeffs = self.options.struct_coeffs
      # Store other parameters
      self.num_mislabel_struct_coeffs = len(self.mislabel_coeffs)
      self.to_tune_mislabel_struct_coeffs = False
      # Check if the lengths are the same
      if len(self.mislabel_coeffs) != len(self.struct_coeffs):
        raise ValueError('Length of mislabel and structural coefficients must be same.' +
                         'Given: mislabel: %s (%d), struct:%s (%d).'%(
                         str(self.options.mislabel_coeffs), len(self.mislabel_coeffs),
                         str(self.options.struct_coeffs), len(self.struct_coeffs)))
    else:
      raise ValueError('Bad format for mislabel and structural coefficients.' +
                       'Given: mislabel: %s, struct:%s.'%(
                      str(self.options.mislabel_coeffs), str(self.options.struct_coeffs)))

  def _get_nn_gp_kernel_options(self):
    """ Returns a dictionary of kernel options. """
    to_copy_over_from_self = ['num_mislabel_struct_coeffs', 'mislabel_coeffs',
      'struct_coeffs', 'to_tune_mislabel_struct_coeffs']
    to_copy_over_from_options = ['dist_type', 'lp_power', 'emd_power',
      'non_assignment_penalty']
    kernel_options = {}
    for elem in to_copy_over_from_self:
      kernel_options[elem] = getattr(self, elem)
    for elem in to_copy_over_from_options:
      kernel_options[elem] = getattr(self.options, elem)
    return kernel_options

  def _child_build_gp(self, mean_func, noise_var, gp_cts_hps, gp_dscr_hps=None,
                      other_gp_params=None, build_posterior=True):
    """ Builds the GP from the hyper-parameters. """
    kernel_options = self._get_nn_gp_kernel_options()
    kernel, gp_cts_hps, gp_dscr_hps = get_nn_gp_kernel(self.kernel_type, self.nn_type,
                                    gp_cts_hps, gp_dscr_hps, self.tp_comp, kernel_options)
    ret = NNGP(self.X, self.Y, kernel, mean_func, noise_var, self.nn_type,
               list_of_dists=self.list_of_dists, build_posterior=build_posterior)
    return ret, gp_cts_hps, gp_dscr_hps


# Some APIs for creating the kernel ---------------------------------------------
def get_nn_kernel_from_args(kernel_type, nn_type, kernel_hyperparams,
                            tp_comp, dist_type):
  """ Returns the kernel. """
  if tp_comp is None:
    tp_comp = otmann.get_otmann_distance_computer_from_args(nn_type,
                kernel_hyperparams['non_assignment_penalty'],
                kernel_hyperparams['mislabel_coeffs'],
                kernel_hyperparams['struct_coeffs'],
                dist_type,
                )
  # Now construct the kernel.
  if kernel_type in ['lpemd_prod', 'lp', 'emd']:
    return otmann.DistProdNNKernel(tp_comp,
            kernel_hyperparams['betas'],
            kernel_hyperparams['scale'],
            kernel_hyperparams['powers'])
  elif kernel_type in ['lpemd_sum']:
    return otmann.DistSumNNKernel(tp_comp,
            kernel_hyperparams['alphas'],
            kernel_hyperparams['betas'],
            kernel_hyperparams['powers'])
  else:
    raise ValueError('Unknown kernel_type %s.'%(kernel_type))


def get_nn_gp_kernel(kernel_type, nn_type, gp_cts_hps, gp_dscr_hps, tp_comp,
                     kernel_options):
  """ Constructs and returns the kernel hyperparams. """
  # pylint: disable=unused-argument
  kernel_hyperparams = {}
  gp_cts_hps = gp_cts_hps[:] # create a copy of the list
  # Extract GP hyper-parameters
  # 1. Scale and alphas
  if kernel_type in ['lp', 'emd', 'lpemd_prod']:
    kernel_hyperparams['scale'] = np.exp(gp_cts_hps[0])
    gp_cts_hps = gp_cts_hps[1:]
  elif kernel_type == 'lpemd_sum':
    kernel_hyperparams['alphas'] = np.exp(gp_cts_hps[:2])
    gp_cts_hps = gp_cts_hps[2:]
  # 2. betas
  if kernel_options['dist_type'] in ['lp', 'emd']:
    kernel_hyperparams['betas'] = np.exp(
      gp_cts_hps[:kernel_options['num_mislabel_struct_coeffs']])
    gp_cts_hps = gp_cts_hps[kernel_options['num_mislabel_struct_coeffs']:]
  elif kernel_options['dist_type'] == 'lp-emd':
    kernel_hyperparams['betas'] = np.exp(
      gp_cts_hps[:2 * kernel_options['num_mislabel_struct_coeffs']])
    gp_cts_hps = gp_cts_hps[2 * kernel_options['num_mislabel_struct_coeffs']:]
  # 3 & 4. mislabel and structural coefficents
  if kernel_options['to_tune_mislabel_struct_coeffs']:
    kernel_hyperparams['mislabel_coeffs'] = gp_cts_hps[0]
    kernel_hyperparams['struct_coeffs'] = gp_cts_hps[1]
    gp_cts_hps = gp_cts_hps[2:]
  else:
    kernel_hyperparams['mislabel_coeffs'] = kernel_options['mislabel_coeffs']
    kernel_hyperparams['struct_coeffs'] = kernel_options['struct_coeffs']
  # Some checks
  assert len(gp_cts_hps) == 0
  # Build the remainder of the kernel ---------------------------------
  # - powers
  lp_powers = [kernel_options['lp_power']] * kernel_options['num_mislabel_struct_coeffs']
  emd_powers = [kernel_options['emd_power']] * \
               kernel_options['num_mislabel_struct_coeffs']
  if kernel_options['dist_type'] == 'lp':
    kernel_hyperparams['powers'] = lp_powers
  elif kernel_options['dist_type'] == 'emd':
    kernel_hyperparams['powers'] = emd_powers
  elif kernel_options['dist_type'] == 'lp-emd':
    kernel_hyperparams['powers'] = [j for i in zip(lp_powers, emd_powers) for j in i]
  # - non_assignment_penalty
  kernel_hyperparams['non_assignment_penalty'] = kernel_options['non_assignment_penalty']
  ret = get_nn_kernel_from_args(kernel_type, nn_type, kernel_hyperparams,
                                tp_comp, kernel_options['dist_type'])
  return ret, gp_cts_hps, gp_dscr_hps




"""
  Unit tests for metropolis sampling of discrete distributions.
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import
import unittest
# pylint: disable=invalid-name
# pylint: disable=no-self-use

import numpy as np

# Local imports
from .discrete import DiscreteUniform, Bernoulli
from ..utils.base_test_class import BaseTestClass, execute_tests

class MetropolisDiscreteTestCases(BaseTestClass):
  """ Unit tests for metropolis sampling of discrete distributions """

  def setUp(self):
    """ Sets up unit tests. """
    self.size = 100000
    self.threshold = 0.1

  def _check_sample_sizes(self, samples):
    """ Compares the sample sizes with the size parameter"""
    assert self.size == len(samples)

  def _compute_mean(self, samples):
    """ Computes Mean """
    return np.mean(samples)

  def _compute_variance(self, samples):
    """ Computes Variance """
    return np.var(samples)

  @unittest.skip
  def test_metropolis_discreteuniform(self):
    """ Tests metropolis sampling from Discrete Uniform distribution """
    self.report('Test metropolis sampling of Discrete Uniform Distribution.')
    lower = -5
    upper = 5
    dist = DiscreteUniform(lower, upper)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('metropolis', self.size, np.array([0.1]))
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert (samples >= lower).all() and (samples <= upper).all()
    assert abs(mean - self._compute_mean(samples)) <= self.threshold
    assert abs(var - self._compute_variance(samples)) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

  def test_metropolis_bernoulli(self):
    """ Tests metropolis sampling from Bernoulli distribution """
    self.report('Test metropolis sampling of Bernoulli Distribution.')
    p = 0.6
    dist = Bernoulli(p)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('binarymetropolis', self.size, np.array([0]))
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert (samples >= 0).all() and (samples <= 1).all()
    assert abs(mean - self._compute_mean(samples)) <= self.threshold
    assert abs(var - self._compute_variance(samples)) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

if __name__ == '__main__':
  execute_tests()



"""
  Adapted from pgmpy library: https://github.com/pgmpy/pgmpy
  A collection of methods for sampling from continuous models
  -- kvysyara@andrew.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=no-self-use
# pylint: disable=arguments-differ

from __future__ import division
from math import sqrt

import numpy as np

from .check_functions import _check_1d_array_object
from .base import LeapFrog, BaseSimulateHamiltonianDynamics

class HamiltonianMC(object):
  """
  Class for performing sampling using simple
  Hamiltonian Monte Carlo

  Parameters:
  -----------
  model: Model from which sampling has to be done

  simulate_dynamics: A class to propose future values of
    momentum and position in time by simulating
    Hamiltonian Dynamics

  References
  ----------
  R.Neal. Handbook of Markov Chain Monte Carlo,
  chapter 5: MCMC Using Hamiltonian Dynamics.
  CRC Press, 2011.
  """

  def __init__(self, model, simulate_dynamics=LeapFrog):

    if not issubclass(simulate_dynamics, BaseSimulateHamiltonianDynamics):
      raise TypeError("split_time must be an instance of " +
                      "BaseSimulateHamiltonianDynamics")

    self.model = model
    self.simulate_dynamics = simulate_dynamics
    self.accepted_proposals = 0.0
    self.acceptance_rate = 0

  def _acceptance_prob(self, position, position_bar, momentum, momentum_bar):
    """
    Returns the acceptance probability for given new position(position) and momentum
    """

    # Parameters to help in evaluating Joint distribution P(position, momentum)
    _logp = self.model.logp(position)
    _logp_bar = self.model.logp(position_bar)

    # acceptance_prob = P(position_bar, momentum_bar)/ P(position, momentum)
    if not np.isfinite(_logp_bar):
      potential_change = _logp_bar
    else:
      potential_change = _logp_bar - _logp  # Negative change
    kinetic_change = 0.5 * np.float(np.dot(momentum_bar.T, momentum_bar) -\
                                    np.dot(momentum.T, momentum))

    # acceptance probability
    #return np.exp(potential_change - kinetic_change)
    return potential_change - kinetic_change

  def _get_condition(self, acceptance_prob, a):
    """
    Temporary method to fix issue in numpy 0.12 #852
    """
    #return (1/(acceptance_prob ** a)) > np.power(2.0, -a)
    #return (acceptance_prob ** a) > np.power(2.0, -a)
    return (acceptance_prob * a) > np.log(np.power(2.0, -a))

  def _find_reasonable_stepsize(self, position, stepsize_app=1):
    """
    Method for choosing initial value of stepsize

    References
    -----------
    Matthew D. Hoffman, Andrew Gelman, The No-U-Turn Sampler: Adaptively
    Setting Path Lengths in Hamiltonian Monte Carlo. Journal of
    Machine Learning Research 15 (2014) 1351-1381
    """
    # momentum = N(0, I)
    momentum = np.reshape(np.random.normal(0, 1, len(position)), position.shape)

    # Take a single step in time
    position_bar, momentum_bar, _ = self.simulate_dynamics(self.model, position,\
                                    momentum, stepsize_app).get_proposed_values()

    acceptance_prob = self._acceptance_prob(position, position_bar, momentum,\
                                            momentum_bar)

    # a = 2I[acceptance_prob] -1
    #a = 2*(acceptance_prob > 0.5) - 1
    a = 2*(acceptance_prob > np.log(0.5)) - 1

    condition = self._get_condition(acceptance_prob, a)

    while condition:
      stepsize_app = np.power(2.0, a)*stepsize_app

      position_bar, momentum_bar, _ = self.simulate_dynamics(self.model, position,\
                                      momentum, stepsize_app).get_proposed_values()

      acceptance_prob = self._acceptance_prob(position, position_bar, momentum,\
                                              momentum_bar)

      condition = self._get_condition(acceptance_prob, a)

    return stepsize_app

  def _sample(self, position, trajectory_length, stepsize, lsteps=None):
    """
    Runs a single sampling iteration to return a sample
    """
    # Resampling momentum
    momentum = np.reshape(np.random.normal(0, 1, len(position)), position.shape)

    # position_m here will be the previous sampled value of position
    position_bar, momentum_bar = position.copy(), momentum

    # Number of steps L to simulate dynamics
    if lsteps is None:
      lsteps = int(max(1, round(trajectory_length / stepsize, 0)))

    grad_bar = self.model.grad_logp(position_bar)

    for _ in range(lsteps):
      position_bar, momentum_bar, grad_bar =\
        self.simulate_dynamics(self.model, position_bar, momentum_bar,\
                               stepsize, grad_bar).get_proposed_values()

    acceptance_prob = self._acceptance_prob(position, position_bar, momentum,
                                            momentum_bar)

    # Metropolis acceptance probability
    alpha = 1
    if acceptance_prob < 0:
      alpha = np.exp(acceptance_prob)

    # Accept or reject the new proposed value of position, i.e position_bar
    if np.random.rand() < alpha:
      position = position_bar.copy()
      self.accepted_proposals += 1.0

    return position, alpha

  def sample(self, initial_pos, num_samples, trajectory_length, stepsize=None, burn=1000):
    """
    Method to return samples using Hamiltonian Monte Carlo

    Parameters
    ----------
    initial_pos: A 1d array like object
      Vector representing values of parameter position, the starting
      state in markov chain.

    num_samples: int
      Number of samples to be generated

    trajectory_length: int or float
      Target trajectory length, stepsize * number of steps(L),
      where L is the number of steps taken per HMC iteration,
      and stepsize is step size for splitting time method.

    stepsize: float, defaults to None
      The stepsize for proposing new values of position and momentum in simulate_dynamics
      If None, then will be choosen suitably

    burn: int, defaults to 1000
      Number of samples to ignore in the beginning
    """


    self.accepted_proposals = 1.0
    initial_pos = _check_1d_array_object(initial_pos, 'initial_pos')

    if stepsize is None:
      stepsize = self._find_reasonable_stepsize(initial_pos)

    #samples = np.zeros(num_samples, dtype=types).view(np.recarray)
    samples = np.zeros(num_samples, dtype=float)

    # Assigning after converting into tuple because value was being
    # changed after assignment. Reason for this is unknown
    position_m = initial_pos

    lsteps = int(max(1, round(trajectory_length / stepsize, 0)))
    for i in range(1, num_samples+burn+1):

      # Genrating sample
      position_m, _ = self._sample(position_m, trajectory_length, stepsize, lsteps)
      # ignore first few samples
      if i >= burn:
        samples[i-burn-1] = position_m

    self.acceptance_rate = self.accepted_proposals / num_samples

    return samples

class HamiltonianMCDA(HamiltonianMC):
  """
  Class for performing sampling in Continuous model
  using Hamiltonian Monte Carlo with dual averaging for
  adaptaion of parameter stepsize.

  Parameters:
  -----------
  model: Model from which sampling has to be done

  simulate_dynamics: Class to propose future states of position
    and momentum in time by simulating HamiltonianDynamics

  delta: float (in between 0 and 1), defaults to 0.65
    The target HMC acceptance probability

  References
  -----------
  Matthew D. Hoffman, Andrew Gelman, The No-U-Turn Sampler: Adaptively
  Setting Path Lengths in Hamiltonian Monte Carlo. Journal of
  Machine Learning Research 15 (2014) 1351-1381
  """

  def __init__(self, model, simulate_dynamics=LeapFrog, delta=0.65):

    if not isinstance(delta, float) or delta > 1.0 or delta < 0.0:
      raise ValueError("delta should be a floating value in between 0 and 1")

    self.delta = delta

    super(HamiltonianMCDA, self).__init__(model=model,\
                                          simulate_dynamics=simulate_dynamics)

  def _adapt_params(self, stepsize, stepsize_bar, h_bar, mu, index_i, alpha, n_alpha=1):
    """
    Run tha adaptation for stepsize for better proposals of position
    """
    gamma = 0.05  # free parameter that controls the amount of shrinkage towards mu
    t0 = 10.0  # free parameter that stabilizes the initial iterations of the algorithm
    kappa = 0.75
    # See equation (6) section 3.2.1 for details

    estimate = 1.0 / (index_i + t0)
    h_bar = (1 - estimate) * h_bar + estimate * (self.delta - (alpha/n_alpha))

    stepsize = np.exp(mu - ((sqrt(index_i)/gamma)*h_bar))
    i_kappa = index_i ** (-kappa)
    stepsize_bar = np.exp(i_kappa*np.log(stepsize) + (1 - i_kappa)*np.log(stepsize_bar))

    return stepsize, stepsize_bar, h_bar

  def sample(self, initial_pos, num_adapt, num_samples, trajectory_length, stepsize=None,
             burn=1000):
    """
    Method to return samples using Hamiltonian Monte Carlo

    Parameters
    ----------
    initial_pos: A 1d array like object
      Vector representing values of parameter position, the starting
      state in markov chain.

    num_adapt: int
      The number of interations to run the adaptation of stepsize

    num_samples: int
      Number of samples to be generated

    trajectory_length: int or float
      Target trajectory length, stepsize * number of steps(L),
      where L is the number of steps taken per HMC iteration,
      and stepsize is step size for splitting time method.

    stepsize: float , defaults to None
      The stepsize for proposing new values of position and momentum in simulate_dynamics
      If None, then will be choosen suitably

    burn: int, defaults to 1000
      Number of samples to ignore in the beginning
    """

    self.accepted_proposals = 1.0

    initial_pos = _check_1d_array_object(initial_pos, 'initial_pos')

    if stepsize is None:
      stepsize = self._find_reasonable_stepsize(initial_pos)

    if num_adapt <= 1:  # Return samples genrated using Simple HMC algorithm
      return HamiltonianMC.sample(self, initial_pos, num_samples, trajectory_length,
                                  stepsize)

    # stepsize is epsilon
    # freely chosen point, after each iteration xt(/position) is shrunk towards it
    mu = np.log(10.0 * stepsize)
    # log(10 * stepsize) large values to save computation
    # stepsize_bar is epsilon_bar
    stepsize_bar = 1.0
    h_bar = 0.0
    # See equation (6) section 3.2.1 for details

    #samples = np.zeros(num_samples, dtype=types).view(np.recarray)
    samples = np.zeros(num_samples, dtype=float)
    position_m = initial_pos

    for i in range(1, num_samples+burn+1):

      # Genrating sample
      position_m, alpha = self._sample(position_m, trajectory_length, stepsize)
      # ignore first few samples
      if i >= burn:
        samples[i-burn-1] = position_m

      # Adaptation of stepsize till num_adapt iterations
      if i <= num_adapt:
        stepsize, stepsize_bar, h_bar = self._adapt_params(stepsize, stepsize_bar, h_bar,
                                                           mu, i, alpha)
      else:
        stepsize = stepsize_bar

    self.acceptance_rate = self.accepted_proposals / num_samples

    return samples



"""
  Class for joint probability distribution of independent random variables
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import
# pylint: disable=invalid-name
# pylint: disable=abstract-method

import numpy as np

# Local imports
from .distribution import Distribution

class JointDistribution(Distribution):
  """ Independent Joint Distribution """

  def __init__(self, rv_s):
    super(JointDistribution, self).__init__()
    self.rv_s = rv_s

  def get_rv_s(self):
    """ Returns the list of random variables """
    return self.rv_s

  def draw_samples(self, method, size=None, init_sample=None, *args):
    """ Returns samples drawn from distribution. """
    if method == 'random':
      return self.draw_random_samples(size)
    elif method == 'nuts':
      return self.draw_samples_nuts(size, init_sample, *args)
    elif method == 'slice':
      return self.draw_samples_slice(size, init_sample, *args)
    elif method == 'metropolis':
      return self.draw_samples_metropolis(size, init_sample, *args)

  def draw_random_samples(self, size=None):
    """ Returns random samples drawn from distribution. """
    samples = np.zeros((len(self.rv_s), size))
    for i, rv in enumerate(self.rv_s):
      samples[i] = rv.draw_samples('random', size)
    return samples

  def draw_samples_slice(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using Slice sampler"""
    samples = np.zeros((len(self.rv_s), size))
    for i, rv in enumerate(self.rv_s):
      samples[i] = rv.draw_samples('slice', size, init_sample, *args)
    return samples

  def draw_samples_nuts(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using NUTS sampler"""
    samples = np.zeros((len(self.rv_s), size))
    for i, rv in enumerate(self.rv_s):
      samples[i] = rv.draw_samples('nuts', size, init_sample, *args)
    return samples

  def draw_samples_metropolis(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using Metropolis sampler"""
    samples = np.zeros((len(self.rv_s), size))
    for i, rv in enumerate(self.rv_s):
      samples[i] = rv.draw_samples('metropolis', size, init_sample, *args)
    return samples



"""
  Test cases for functions/classes in neural_network.py
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=no-name-in-module
# pylint: disable=invalid-name

import numpy as np
# Local imports
from . import neural_network
from .nn_examples import generate_cnn_architectures, generate_mlp_architectures
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.graph_utils import apsp_floyd_warshall_costs


# Test cases for NeuralNetwork class -------------------------------------------------
class NeuralNetworkTestCase(BaseTestClass):
  """ Unit tests for the NeuralNetwork test class. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(NeuralNetworkTestCase, self).__init__(*args, **kwargs)
    self.cnns = generate_cnn_architectures()
    self.mlps_reg = generate_mlp_architectures('reg')
    self.mlps_class = generate_mlp_architectures('class')

  def _print_path_lengths(self, nns, fwd_or_bkwd, path_length_types):
    """ Prints the path length types. """
    # pylint: disable=protected-access
    # disabling this because this property should be private but printing this out during
    # a test might help catch some bugs.
    for idx, nn in enumerate(nns):
      self.report('%s-%d:: #layers=%d, #edges=%d.'%(nn.nn_class, idx, nn.num_layers,
                                                    nn.conn_mat.sum()), 'test_result')
      if fwd_or_bkwd == 'fwd':
        curr_path_lengths = nn._fwd_dists_to_op
      elif fwd_or_bkwd == 'bkwd':
        curr_path_lengths = nn._bkwd_dists_to_ip
      else:
        curr_path_lengths = None
      # Print them out
      for plt in path_length_types:
        self.report('  %s[%s]: %s.'%(fwd_or_bkwd, plt, curr_path_lengths[plt]),
                    'test_result')

  def test_cnn_instantiation(self):
    """ Tests instantiation. """
    self.report('Testing instantiation of cnns.')
    path_length_types = ['all-shortest', 'all-longest', 'all-rw',
                         'conv-shortest', 'conv-longest', 'conv-rw',
                         'pool-shortest', 'pool-longest', 'pool-rw',
                         'fc-shortest', 'fc-longest', 'fc-rw',
                        ]
    self._print_path_lengths(self.cnns, 'bkwd', path_length_types)
    self._print_path_lengths(self.cnns, 'fwd', path_length_types)

  def test_mlp_instantiation(self):
    """ Tests instantiation. """
    path_length_types = ['all-shortest', 'all-longest', 'all-rw',
                         'sigmoid-shortest', 'sigmoid-longest', 'sigmoid-rw',
                         'rectifier-shortest', 'rectifier-longest', 'rectifier-rw',
                        ]
    self.report('Testing instantiation of mlps - regression.')
    self._print_path_lengths(self.mlps_reg, 'bkwd', path_length_types)
    self._print_path_lengths(self.mlps_reg, 'fwd', path_length_types)
    self.report('Testing instantiation of mlps - classification.')
    self._print_path_lengths(self.mlps_class, 'bkwd', path_length_types)
    self._print_path_lengths(self.mlps_class, 'fwd', path_length_types)

  def test_bkwd_ip_fwd_op_dists_of_all_layers(self):
    """ Tests computation and retrieval of the distances to all layers. """
    self.report('Testing computation/retrieval of backward and forward distances.')
    for cnn in self.cnns:
      all_label_order = ['all-shortest', 'conv-shortest', 'pool-shortest']
      np.random.shuffle(all_label_order)
      # First comptue the distances and check if the sizes are ok.
      curr_bkwd_to_ip_comp, curr_fwd_to_op_comp = \
        cnn.get_bkwd_ip_fwd_op_dists_of_all_layers(all_label_order)
      assert curr_fwd_to_op_comp.shape == (cnn.num_layers, len(all_label_order))
      assert curr_bkwd_to_ip_comp.shape == (cnn.num_layers, len(all_label_order))
      # Now check for values - we will use floyd warshall to compute the distances here.
      ip_idx = cnn.get_ip_layer_idx()
      op_idx = cnn.get_op_layer_idx()
      curr_edge_wts = cnn.get_edge_weights_from_conn_mat()
      curr_fwd_to_op = []
      curr_bkwd_to_ip = []
      # compute the all pairs distances
      for ll in all_label_order:
        if ll == 'all-shortest':
          curr_fwd_to_op.append(
            apsp_floyd_warshall_costs(curr_edge_wts)[:, op_idx])
          curr_bkwd_to_ip.append(
            apsp_floyd_warshall_costs(curr_edge_wts.T)[:, ip_idx])
        else:
          dist_type = ll.split('-')[0]
          ll_edge_wts = cnn.get_layer_or_group_edge_weights_from_edge_weights(
                          curr_edge_wts.T, dist_type)
          ll_edge_wtsT = cnn.get_layer_or_group_edge_weights_from_edge_weights(
                           curr_edge_wts, dist_type)
          curr_bkwd_to_ip.append(
            apsp_floyd_warshall_costs(ll_edge_wtsT)[ip_idx, :])
          curr_fwd_to_op.append(
            apsp_floyd_warshall_costs(ll_edge_wts)[op_idx, :])
      curr_fwd_to_op = np.array(curr_fwd_to_op).T
      curr_bkwd_to_ip = np.array(curr_bkwd_to_ip).T
#       assert np.all(curr_bkwd_to_ip == curr_bkwd_to_ip_comp)
#       assert np.all(curr_fwd_to_op == curr_fwd_to_op_comp)


# Test cases for general NN functions -------------------------------------------------
class GeneralNNFunctionsTestCase(BaseTestClass):
  """ Unit tests for general NN functions. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(GeneralNNFunctionsTestCase, self).__init__(*args, **kwargs)

  def test_get_labels(self):
    """ Tests labels generated for a CNN and MLP. """
    self.report('Testing the labels for CNN and MLP.')
    cnn_true_labels_7 = sorted(['ip', 'op', 'fc', 'max-pool', 'avg-pool', 'softmax',
                                'res3', 'res5', 'res7', 'conv3', 'conv5', 'conv7'])
    cnn_true_labels_3 = sorted(['ip', 'op', 'fc', 'max-pool', 'avg-pool', 'softmax',
                                'res3', 'conv3'])
    mlp_true_labels = sorted(['ip', 'op', 'relu', 'linear'])
    cnn_ret_labels_7 = sorted(neural_network.get_cnn_layer_labels(7))
    cnn_ret_labels_3 = sorted(neural_network.get_cnn_layer_labels(3))
    mlp_ret_labels = sorted(neural_network.get_mlp_layer_labels('reg', ['relu']))
    assert cnn_true_labels_7 == cnn_ret_labels_7
    assert cnn_true_labels_3 == cnn_ret_labels_3
    assert mlp_true_labels == mlp_ret_labels


if __name__ == '__main__':
  execute_tests()




"""
  APIs for Multi-objective Optimisation.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=maybe-no-member


from .api_utils import preprocess_arguments, get_worker_manager_from_capital_type, \
                       load_options_for_method
from ..exd.experiment_caller import EuclideanMultiFunctionCaller, CPMultiFunctionCaller
from ..exd.cp_domain_utils import get_raw_from_processed_via_config
from ..opt.multiobjective_gp_bandit import multiobjective_gpb_from_multi_func_caller
from ..opt.random_multiobjective_optimiser import \
                                 random_multiobjective_optimisation_from_multi_func_caller


def multiobjective_maximise_functions(funcs, domain, max_capital,
                                      capital_type='num_evals', opt_method='bo',
                                      config=None, options=None, reporter='default'):
  """
    Co-optimises the functions 'funcs' over the domain 'domain'.
    Inputs:
      funcs: The functions to be co-optimised (maximised).
      domain: The domain over which the function should be maximised, should be an
              instance of the Domain class in exd/domains.py.
              If domain is a list of the form [[l1, u1], [l2, u2], ...] where li < ui,
              then we will create a Euclidean domain with lower bounds li and upper bounds
              ui along each dimension.
      max_capital: The maximum capital (time budget or number of evaluations) available
                   for optimisation.
      capital_type: The type of capital. Should be one of 'return_value' or 'realtime'.
                    Default is return_value which indicates we will use the value returned
                    by fidel_cost_func. If realtime, we will use wall clock time.
      opt_method: The method used for optimisation. Could be one of bo or rand.
                  Default is bo. bo: Bayesian optimisation, rand: Random search.
      config: Contains configuration parameters that are typically returned by
              exd.cp_domain_utils.load_config_file. config can be None only if domain
              is a EuclideanDomain object.
      options: Additional hyper-parameters for optimisation, as a namespace.
      reporter: A stream to print progress made during optimisation, or one of the
                following strings 'default', 'silent'. If 'silent', then it suppresses
                all outputs. If 'default', writes to stdout.
      * Alternatively, domain could be None if config is either a path_name to a
        configuration file or has configuration parameters.
    Returns:
      pareto_values: The pareto optimal values found during the optimisation procdure.
      pareto_points: The corresponding pareto optimum points in the domain.
      history: A record of the optimisation procedure which include the point evaluated
               and the values at each time step.
  """
  # Preprocess domain and config arguments
  raw_funcs = funcs
  domain, funcs, config, _ = preprocess_arguments(domain, funcs, config)
  # Load arguments depending on domain type
  if domain.get_type() == 'euclidean':
    multi_func_caller = EuclideanMultiFunctionCaller(funcs, domain, vectorised=False)
  else:
    multi_func_caller = CPMultiFunctionCaller(funcs, domain, raw_funcs=raw_funcs,
                          domain_orderings=config.domain_orderings)
  # load options
  options = load_options_for_method(opt_method, 'moo', domain, capital_type, options)
  # Create worker manager and function caller
  worker_manager = get_worker_manager_from_capital_type(capital_type)

  # Select method here -------------------------------------------------------------------
  if opt_method == 'bo':
    pareto_values, pareto_points, history = multiobjective_gpb_from_multi_func_caller(
                                        multi_func_caller, worker_manager, max_capital,
                                        is_mf=False, options=options, reporter=reporter)
  elif opt_method == 'rand':
    pareto_values, pareto_points, history = \
      random_multiobjective_optimisation_from_multi_func_caller(multi_func_caller,
        worker_manager, max_capital, mode='asy', options=options, reporter=reporter)

  # Post processing
  if domain.get_type() == 'euclidean' and config is None:
    pareto_points = [multi_func_caller.get_raw_domain_coords(pt) for pt in pareto_points]
    history.query_points = [multi_func_caller.get_raw_domain_coords(pt)
                            for pt in history.query_points]
  else:
    pareto_points = [get_raw_from_processed_via_config(pt, config) for
                     pt in pareto_points]
    history.query_points_raw = [get_raw_from_processed_via_config(pt, config)
                                for pt in history.query_points]
  return pareto_values, pareto_points, history


def multiobjective_minimise_functions(funcs, *args, **kwargs):
  """
    Minimises the functions funcs over domain domain. See
    multiobjective_maximise_functions for a description of the arguments. All arguments
    are the same except for funcs, which should now be minimised.
  """
  def _get_func_to_max(_func):
    """ Returns a function to be maximised. """
    return lambda x: - _func(x)
  funcs_to_max = [_get_func_to_max(f) for f in funcs]
  pareto_max_values, pareto_points, history = multiobjective_maximise_functions(
                                                funcs_to_max, *args, **kwargs)
  # Post process history
  def _negate_array(arr):
    """ Negates an array. """
    return [-elem for elem in arr]
  pareto_min_values = _negate_array(pareto_max_values)
  history.query_vals = [_negate_array(qv) for qv in history.query_vals]
  history.curr_pareto_vals = [_negate_array(cpv) for cpv in history.curr_pareto_vals]
  history.curr_true_pareto_vals = [_negate_array(ctpv) for ctpv in
                                   history.curr_true_pareto_vals]
  return pareto_min_values, pareto_points, history


# Alternative spelling
multiobjective_maximize_functions = multiobjective_maximise_functions
multiobjective_minimize_functions = multiobjective_minimise_functions




"""
  Test cases for functions in graph_utils.py
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=import-error
# pylint: disable=invalid-name

import numpy as np
from scipy.sparse import dok_matrix
from time import clock
# Local
from . import graph_utils
from .base_test_class import BaseTestClass, execute_tests


ERR_TOL_FRAC = 1e-5


class GraphUtilsTestCase(BaseTestClass):
  """ Unit test class for ancillary utilities. """

  def setUp(self):
    """ Sets up attributes. """
    self.num_vertices_values = [4, 10, 50, 100]

  def test_apsp_algos(self):
    """ Tests the all pairs shortest pair problems. """
    # Todo: write code to compute computation time.
    self.report('Testing All Pairs Shortest Paths for n=%s.'%(
                str(self.num_vertices_values)))
    for num_vert_val in self.num_vertices_values:
      # Floyd warshall
      A = graph_utils.construct_random_adjacency_matrix(num_vert_val, avg_edge_frac=0.4,
                                                        directed=True)
      num_edges = A.sum()
      fw_start_time = clock()
      fw_dist = graph_utils.apsp_floyd_warshall(A) # call floyd_warshall
      fw_end_time = clock()
      fw_time_taken = fw_end_time - fw_start_time
      # Dijkstra's
      dj_start_time = clock()
      dj_dist = graph_utils.apsp_dijkstra(A) # call dijkstra's
      dj_end_time = clock()
      dj_time_taken = dj_end_time - dj_start_time
      # Report times taken
      self.report('Time taken for n=%d, m=%d: FW: %0.4f, Dijkstra\'s: %0.4f.'%(
                  A.shape[0], num_edges, fw_time_taken, dj_time_taken))
      # Test if both are equal
      fw_infties = fw_dist == np.inf
      dj_infties = dj_dist == np.inf
      assert np.all(fw_infties == dj_infties)
      # Zer out the infinities and compare the norms
      fw_dist[fw_infties] = 0
      dj_dist[dj_infties] = 0
      error_tol = ERR_TOL_FRAC * np.linalg.norm(dj_dist + fw_dist)
      assert np.linalg.norm(fw_dist - dj_dist) < error_tol

  def test_kahn_topoligical_sort(self):
    """ Tests topological sorting. """
    self.report('Testing topological sort.')
    A = dok_matrix((6, 6))
    A[0, 2] = 1
    A[1, 2] = 1
    A[1, 3] = 1
    A[2, 3] = 1
    A[2, 4] = 1
    A[4, 3] = 1
    A[5, 0] = 1
    A[5, 1] = 1
    sorted_order, has_cycles = graph_utils.kahn_topological_sort(A, 5)
    assert has_cycles == False
    assert sorted_order == [5, 0, 1, 2, 4, 3] or sorted_order == [5, 1, 0, 2, 4, 3]
    # Now create a graph with cycles
    A[3, 0] = 1
    sorted_order, has_cycles = graph_utils.kahn_topological_sort(A, 5)
    assert has_cycles == True

if __name__ == '__main__':
  execute_tests()




"""
  Implements some standard synthetic functions in Euclidean spaces.
  -- kandasamy@cs.cmu.edu
"""
from __future__ import division

# pylint: disable=invalid-name

import numpy as np
# Local imports
from .general_utils import map_to_cube
from ..exd.experiment_caller import EuclideanFunctionCaller


# Hartmann Functions ---------------------------------------------------------------------
def hartmann(x, alpha, A, P, max_val=np.inf):
  """ Computes the hartmann function for any given A and P. """
  log_sum_terms = (A * (P - x)**2).sum(axis=1)
  return min(max_val, alpha.dot(np.exp(-log_sum_terms)))

def _get_hartmann_data(domain_dim):
  """ Returns A and P for the 3D hartmann function. """
  # pylint: disable=bad-whitespace
  if domain_dim == 3: # 3D hartmann function
    A = np.array([[3.0, 10, 30],
                  [0.1, 10, 35],
                  [3.0, 10, 30],
                  [0.1, 10, 35]], dtype=np.float64)
    P = 1e-4 * np.array([[3689, 1170, 2673],
                         [4699, 4387, 7470],
                         [1091, 8732, 5547],
                         [ 381, 5743, 8828]], dtype=np.float64)
    alpha = np.array([1.0, 1.2, 3.0, 3.2])
    domain = [[0, 1]] * 3
    opt_pt = np.array([0.114614, 0.555649, 0.852547])
    max_val = 3.86278
  elif domain_dim == 6: # 6D hartmann function
    A = np.array([[  10,   3,   17, 3.5, 1.7,  8],
                  [0.05,  10,   17, 0.1,   8, 14],
                  [   3, 3.5,  1.7,  10,  17,  8],
                  [  17,   8, 0.05,  10, 0.1, 14]], dtype=np.float64)
    P = 1e-4 * np.array([[1312, 1696, 5569,  124, 8283, 5886],
                         [2329, 4135, 8307, 3736, 1004, 9991],
                         [2348, 1451, 3522, 2883, 3047, 6650],
                         [4047, 8828, 8732, 5743, 1091,  381]], dtype=np.float64)
    alpha = np.array([1.0, 1.2, 3.0, 3.2])
    domain = [[0, 1]] * 6
    opt_pt = np.array([0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573])
    max_val = 3.322368
  else:
    raise NotImplementedError('Only implemented in 3 and 6 dimensions.')
  return A, P, alpha, opt_pt, domain, max_val

def get_mf_hartmann_function_data(fidel_dim, domain_dim):
  """ Returns a function f(z, x). z refers to the fidelity and x is the point in the
      domain. """
  A, P, alpha, opt_pt, domain_bounds, max_val = _get_hartmann_data(domain_dim)
  # This is how much we will perturb the alphas
  delta = np.array([0.1] * fidel_dim + [0] * (4-fidel_dim))
  # Define a wrapper for the mf objective
  def mf_hart_obj(z, x):
    """ Wrapper for the mf hartmann objective. z is fidelity and x is domain. """
    assert len(z) == fidel_dim
    z_extended = np.append(z, [0] * (4-fidel_dim))
    alpha_z = alpha - (1 - z_extended) * delta
    return hartmann(x, alpha_z, A, P, max_val)
  # Define a wrapper for the sf objective
  def hart_obj(x):
    """ Wrapper for the hartmann objective. z is fidelity and x is domain. """
    return hartmann(x, alpha, A, P, max_val)
  # Define the optimum fidelity and the fidelity bounds
  fidel_to_opt = np.ones(fidel_dim)
  fidel_bounds = [[0, 1]] * fidel_dim
  opt_val = hart_obj(opt_pt)
  return mf_hart_obj, hart_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds


# Currin Exponential Function ------------------------------------------------------------
def currin_exp(x, alpha):
  """ Computes the currin exponential function. """
  x1 = x[0]
  x2 = x[1]
  val_1 = 1 - alpha * np.exp(-1/(2 * x2))
  val_2 = (2300*x1**3 + 1900*x1**2 + 2092*x1 + 60) / (100*x1**3 + 500*x1**2 + 4*x1 + 20)
  return val_1 * val_2

def get_mf_currin_exp_function_data():
  """ Returns the multi-fidelity currin exponential function with d=6 and p=2. """
  opt_val = 13.7986850
  # A wrapper for the MF objective
  def mf_currin_exp_obj(z, x):
    """ Wrapper for the MF currin objective. """
    alpha_z = 1 - 0.1 * z
    return min(opt_val, currin_exp(x, alpha_z))
  fidel_to_opt = np.array([1])
  opt_pt = None
  # A wrapper for the SF objective
  def sf_currin_exp_obj(x):
    """ Wrapper for the single fidelity currin objective. """
    return min(opt_val, currin_exp(x, fidel_to_opt))
  fidel_bounds = np.array([[0, 1]])
  domain_bounds = np.array([[0, 1], [0, 1]])
  return mf_currin_exp_obj, sf_currin_exp_obj, opt_pt, opt_val, fidel_to_opt, \
         fidel_bounds, domain_bounds


# Branin Function ------------------------------------------------------------------------
def branin_function(x, a, b, c, r, s, t):
  """ Computes the Branin function. """
  x1 = x[0]
  x2 = x[1]
  neg_ret = float(a * (x2 - b*x1**2 + c*x1 - r)**2 + s*(1-t)*np.cos(x1) + s)
  return -neg_ret

def branin_function_alpha(x, alpha, a, r, s):
  """ Alternative form for the branin function. """
  return branin_function(x, a, alpha[0], alpha[1], r, s, alpha[2])

def get_mf_branin_function(fidel_dim):
  """ Returns the Branin function as a multifidelity function. """
  a0 = 1
  b0 = 5.1/(4*np.pi**2)
  c0 = 5/np.pi
  r0 = 6
  s0 = 10
  t0 = 1/(8*np.pi)
  alpha = np.array([b0, c0, t0])
  # Define delta
  delta = [0.01, 0.1, -0.005]
  delta = np.array(delta[0:fidel_dim] + [0] * (3 - fidel_dim))
  # A wrapper for the MF objective
  def mf_branin_obj(z, x):
    """ Wrapper for the MF Branin objective. """
    assert len(z) == fidel_dim
    z_extended = np.append(z, [0] * (3-fidel_dim))
    alpha_z = alpha - (1 - z_extended) * delta
    return branin_function_alpha(x, alpha_z, a0, r0, s0)
  # A wrapper for the SF objective
  def sf_branin_obj(x):
    """ Wrapper for the SF Branin objective. """
    return branin_function(x, a0, b0, c0, r0, s0, t0)
  # Other data
  fidel_to_opt = np.ones((fidel_dim))
  fidel_bounds = [[0, 1]] * fidel_dim
  opt_pt = np.array([np.pi, 2.275])
  opt_val = sf_branin_obj(opt_pt)
  domain_bounds = [[-5, 10], [0, 15]]
  return mf_branin_obj, sf_branin_obj, opt_pt, opt_val, fidel_to_opt, \
         fidel_bounds, domain_bounds

# Borehole Function ----------------------------------------------------------------------
def borehole_function(x, z, max_val):
  """ Computes the Bore Hole function. """
  # pylint: disable=bad-whitespace
  rw = x[0]
  r  = x[1]
  Tu = x[2]
  Hu = x[3]
  Tl = x[4]
  Hl = x[5]
  L  = x[6]
  Kw = x[7]
  # Compute high fidelity function
  frac2 = 2*L*Tu/(np.log(r/rw) * rw**2 * Kw)
  f2 = min(max_val, 2 * np.pi * Tu * (Hu - Hl)/(np.log(r/rw) * (1 + frac2 + Tu/Tl)))
  # Compute low fidelity function
  f1 = 5 * Tu * (Hu - Hl)/(np.log(r/rw) * (1.5 + frac2 + Tu/Tl))
  # Compute final output
  return float(f2*z + f1*(1-z))

def get_mf_borehole_function():
  """ Gets the MF BoreHole function. """
  opt_val = 309.523221
  opt_pt = None
  mf_borehole_function = lambda z, x: borehole_function(x, z, opt_val)
  domain_bounds = [[0.05, 0.15],
                   [100, 50000],
                   [63070, 115600],
                   [990, 1110],
                   [63.1, 116],
                   [700, 820],
                   [1120, 1680],
                   [9855, 12045]]
  fidel_bounds = [[0, 1]]
  fidel_to_opt = np.array([1])
  # A wrapper for the single fidelity function
  def sf_borehole_obj(x):
    """ A wrapper for the single fidelity objective. """
    return borehole_function(x, fidel_to_opt, opt_val)
  # return
  return mf_borehole_function, sf_borehole_obj, opt_pt, opt_val, fidel_to_opt, \
         fidel_bounds, domain_bounds

# Park Functions ----------------------------------------------------------------------
def park1(x, max_val):
  """ Computes the park1 function. """
  x1 = x[0]
  x2 = x[1]
  x3 = x[2]
  x4 = x[3]
  ret1 = (x1/2) * (np.sqrt(1 + (x2 + x3**2)*x4/(x1**2)) - 1)
  ret2 = (x1 + 3*x4) * np.exp(1 + np.sin(x3))
  return min(ret1 + ret2, max_val)

def get_mf_park1_function():
  """ Gets the MF Park1 function. """
  opt_val = 25.5872304
  opt_pt = None
  sf_park1_obj = lambda x: park1(x, opt_val)
  domain_bounds = [[0, 1]] * 4
  return None, sf_park1_obj, opt_pt, opt_val, None, None, domain_bounds

def park2(x, max_val):
  """ Comutes the park2 function """
  x1 = x[0]
  x2 = x[1]
  x3 = x[2]
  x4 = x[3]
  ret = (2.0/3.0) * np.exp(x1 + x2) - x4*np.sin(x3) + x3
  return min(ret, max_val)

def get_mf_park2_function():
  """ Gets the MF Park2 function. """
  opt_val = 5.925698
  opt_pt = None
  sf_park2_obj = lambda x: park2(x, opt_val)
  domain_bounds = [[0, 1]] * 4
  return None, sf_park2_obj, opt_pt, opt_val, None, None, domain_bounds

# A cost function for MF evaluations -------------------------------------------
def get_mf_cost_function(fidel_bounds):
  """ Returns the cost function. fidel_bounds are the bounds for the fidelity space. """
  fidel_dim = len(fidel_bounds)
  if fidel_dim == 1:
    fidel_powers = [2]
  elif fidel_dim == 2:
    fidel_powers = [3, 2]
  elif fidel_dim == 3:
    fidel_powers = [3, 2, 1.5]
  else:
    fidel_powers = [3] + list(np.linspace(2, 1.2, fidel_dim-1))
  # Define the normalised
  def _unnorm_cost_function(norm_z):
    """ The cost function with normalised coordinates. """
    return np.power(norm_z, fidel_powers).sum()
  max_unnorm_cost = _unnorm_cost_function(np.ones(fidel_dim))
  fidel_bounds = np.array(fidel_bounds)
  def _norm_cost_function(z):
    """ Normalised cost function. """
    return 0.1 + 0.9 * (_unnorm_cost_function(map_to_cube(np.array(z), fidel_bounds)) /
                        max_unnorm_cost)
  return _norm_cost_function

def get_high_dim_function(domain_dim, group_dim, mf_obj, sf_obj):
  """ Constructs a higher dimensional functions. """
  num_groups = int(domain_dim/group_dim)
  def mf_obj_high_dim(z, x):
    """ Evaluates the higher dimensional function. """
    ret = mf_obj(z, x[0:group_dim])
    for j in range(1, num_groups):
      ret += sf_obj(x[j*group_dim: (j+1)*group_dim])
    return ret
  def sf_obj_high_dim(x):
    """ Evaluates the higher dimensional function. """
    ret = 0
    for j in range(num_groups):
      ret += sf_obj(x[j*group_dim: (j+1)*group_dim])
    return ret
  return mf_obj_high_dim, sf_obj_high_dim, num_groups

def get_high_dim_function_data(func_name, fidel_dim=None):
  """ Gets a high dimensional function data from the description. """
  fidel_dim_to_pass = 1 if fidel_dim is None else fidel_dim
  segments = func_name.split('-')
  domain_dim = int(segments[1])
  mf_obj, sf_obj, _, _, fidel_to_opt, fidel_bounds, domain_bounds = \
                               get_function_data(segments[0], fidel_dim=fidel_dim_to_pass)
  group_dim = len(domain_bounds)
  mf_obj_high_dim, sf_obj_high_dim, num_groups = get_high_dim_function(domain_dim, \
                                                               group_dim, mf_obj, sf_obj)
  high_d_domain_bounds = np.tile(np.array(domain_bounds).T, num_groups+1).T[0:domain_dim]

  return mf_obj_high_dim, sf_obj_high_dim, None, None, fidel_to_opt, \
         fidel_bounds, high_d_domain_bounds

def get_function_data(func_name, domain_dim=None, fidel_dim=None, noise_type='no_noise',
                      noise_scale=None):
  """ Gets a function data from the description. """
  # pylint: disable=too-many-branches
  fidel_dim_to_pass = 1 if fidel_dim is None else fidel_dim
  if func_name == 'hartmann':
    if domain_dim is None:
      domain_dim = 6
    if fidel_dim is not None and fidel_dim > 4:
      raise ValueError(('For the hartmann functions, fidel_dim has to be 4 or less. ' +
                        'Given: %s.')%(fidel_dim))
    mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds = \
      get_mf_hartmann_function_data(fidel_dim_to_pass, domain_dim)
  elif func_name == 'hartmann3':
    return get_function_data('hartmann', 3, fidel_dim, noise_type, noise_scale)
  elif func_name == 'hartmann6':
    return get_function_data('hartmann', 6, fidel_dim, noise_type, noise_scale)
  elif func_name == 'branin':
    if domain_dim != 2 and domain_dim is not None:
      raise ValueError(('For the branin function, domain_dim has to be 2 or None. ' +
                        'Given: %s.')%(domain_dim))
    if fidel_dim is not None and fidel_dim > 3:
      raise ValueError(('For the branin function, fidel_dim has to be 3 or less. ' +
                        'Given: %s.')%(fidel_dim))
    mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds = \
      get_mf_branin_function(fidel_dim_to_pass)
  elif func_name == 'borehole':
    if domain_dim != 8 and domain_dim is not None:
      raise ValueError(('For the borehole function, domain_dim has to be 8 or None. ' +
                        'Given: %s.')%(domain_dim))
    if fidel_dim != 1 and fidel_dim is not None:
      raise ValueError(('For the borehole function, fidel_dim has to be 1 or None. ' +
                        'Given: %s.')%(fidel_dim))
    mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds = \
      get_mf_borehole_function()
  elif func_name == 'park1':
    if domain_dim != 4 and domain_dim is not None:
      raise ValueError(('For the park1 function, domain_dim has to be 4 or None. ' +
                        'Given: %s.')%(domain_dim))
    mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds = \
      get_mf_park1_function()
  elif func_name == 'park2':
    if domain_dim != 4 and domain_dim is not None:
      raise ValueError(('For the park2 function, domain_dim has to be 4 or None. ' +
                        'Given: %s.')%(domain_dim))
    mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds = \
      get_mf_park2_function()
  else:
    raise ValueError('Unknwon func_name: %s.'%(func_name))

  return mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds

# An API that returns a function caller from the description --------------------------
def get_syn_func_caller(func_name, domain_dim=None, fidel_dim=None,
                        noise_type='no_noise', noise_scale=None,
                        to_normalise_domain=True):
  """ Returns a FunctionCaller object from the function name. """
  func_name = func_name.lower()
  funcs = ['hartmann', 'hartmann6', 'hartmann3', 'branin', 'borehole', 'park1', 'park2']
  if func_name in funcs:
    mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds = \
      get_function_data(func_name, domain_dim, fidel_dim, noise_type, noise_scale)
  else:
    mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds = \
      get_high_dim_function_data(func_name, fidel_dim)

  # Now return
  if fidel_dim is None:
    return EuclideanFunctionCaller(sf_obj, domain_bounds, descr=func_name, \
                          vectorised=False, to_normalise_domain=to_normalise_domain, \
                          raw_argmax=opt_pt, maxval=opt_val, noise_type=noise_type, \
                          noise_scale=noise_scale)
  else:
    fidel_cost_func = get_mf_cost_function(fidel_bounds)
    return EuclideanFunctionCaller(mf_obj, raw_domain=domain_bounds, descr=func_name, \
                          vectorised=False, to_normalise_domain=to_normalise_domain, \
                          raw_argmax=opt_pt, maxval=opt_val, noise_type=noise_type, \
                          noise_scale=noise_scale, fidel_cost_func=fidel_cost_func, \
                          raw_fidel_space=fidel_bounds, raw_fidel_to_opt=fidel_to_opt)

# An API that returns function data from the description --------------------------
def get_syn_function(func_name, noise_type='no_noise', noise_scale=None):
  """ Returns a synthetic function from the function name along with its
      optimum point, optimum value and the domain_bounds. sf_obj is the
      concerned synthetic function.
  """
  func_name = func_name.lower()
  funcs = ['hartmann', 'hartmann6', 'hartmann3', 'branin', 'borehole', 'park1', 'park2']
  if func_name in funcs:
    mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds = \
      get_function_data(func_name, noise_type=noise_type, noise_scale=noise_scale)
  else:
    mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds = \
      get_high_dim_function_data(func_name)

  return mf_obj, sf_obj, opt_pt, opt_val, fidel_to_opt, fidel_bounds, domain_bounds



"""
  Some utilities when optimising over neural networks.
  --kandasamy@cs.cmu.edu
"""
# pylint: disable=no-member

import numpy as np
# Local
from . import nn_examples
from ..exd.experiment_caller import FunctionCaller


class NNFunctionCaller(FunctionCaller):
  """ Mostly just a place holder class for a Neural network function caller. """
  # pylint: disable=abstract-method
  pass


# Initial pool of CNNs and MLPs ==========================================================
def get_initial_cnn_pool():
  """ Returns the initial pool for CNNs. """
  vgg_nets = [nn_examples.get_vgg_net(1),
              nn_examples.get_vgg_net(2),
              nn_examples.get_vgg_net(3),
              nn_examples.get_vgg_net(4)]
  blocked_cnns = [nn_examples.get_blocked_cnn(3, 1, 1), # 3
                  nn_examples.get_blocked_cnn(3, 2, 1), # 6
                  nn_examples.get_blocked_cnn(3, 3, 1), # 9
                  nn_examples.get_blocked_cnn(3, 4, 1), # 12
                  nn_examples.get_blocked_cnn(3, 5, 1), # 15
                  nn_examples.get_blocked_cnn(4, 4, 1), # 16
                 ]
  resnet_cnns = []
  multidepth_cnns = []
  return vgg_nets + blocked_cnns + resnet_cnns + multidepth_cnns

def get_initial_mlp_pool(class_or_reg):
  """ Returns the initial pool of MLPs. """
  blocked_mlps = [nn_examples.get_blocked_mlp(class_or_reg, 3, 2), # 6
                  nn_examples.get_blocked_mlp(class_or_reg, 4, 2), # 8
                  nn_examples.get_blocked_mlp(class_or_reg, 5, 2), # 10
                  nn_examples.get_blocked_mlp(class_or_reg, 3, 4), # 12
                  nn_examples.get_blocked_mlp(class_or_reg, 6, 2), # 12
                  nn_examples.get_blocked_mlp(class_or_reg, 8, 2), # 16
                  nn_examples.get_blocked_mlp(class_or_reg, 6, 3), # 18
                  nn_examples.get_blocked_mlp(class_or_reg, 10, 2), #20
                  nn_examples.get_blocked_mlp(class_or_reg, 4, 6), #24
                  nn_examples.get_blocked_mlp(class_or_reg, 8, 3), #24
                 ]
  multidepth_mlps = []
  return blocked_mlps + multidepth_mlps


def get_initial_pool(nn_type):
  """ Returns the initial pool from the neural network type. """
  if nn_type.startswith('cnn'):
    return get_initial_cnn_pool()
  elif nn_type.startswith('mlp'):
    return get_initial_mlp_pool(nn_type[4:])


def random_sample_from_nn_domain(nn_type, num_samples, sample_type='rand',
                                 nn_constraint_checker=None):
  """ Randomly samples num_samples neural networks. """
  initial_nns = get_initial_pool(nn_type)
  np.random.shuffle(initial_nns)
  if sample_type == 'ff_only':
    if num_samples <= len(initial_nns):
      return initial_nns[:num_samples]
    else:
      num_add_nns = num_samples - len(initial_nns)
      additional_nns = np.random.choice(initial_nns, size=num_add_nns, replace=True)
      return initial_nns + additional_nns
  elif sample_type == 'rand':
    ret = nn_examples.generate_many_neural_networks(nn_type, num_samples, initial_nns,
                                                    nn_constraint_checker)
    if len(ret) < num_samples:
      additional_nns = np.random.choice(ret, size=num_samples-len(ret), replace=True)
      ret += additional_nns
    np.random.shuffle(ret)
    return ret
  else:
    raise ValueError('Unknown sample type %s.'%(sample_type))




"""
  Unit tests for the GA Optimiser on Cartesian spaces.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=abstract-class-little-used

# Local
from . import cp_ga_optimiser
from .unittest_cp_random_optimiser import CPOptimiserBaseTestCase
from ..utils.base_test_class import BaseTestClass, execute_tests


class CPGAOPtimiserTestCaseDefinitions(object):
  """ Definitions of unit tests for GA optimiser on cartesian product spaces. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager, options, reporter):
    """ Instantiate the optimiser. """
    return cp_ga_optimiser.CPGAOptimiser(func_caller, worker_manager, options=options,
                                         reporter=reporter)

  @classmethod
  def _run_optimiser(cls, prob_funcs, domain_config_file, worker_manager, max_capital,
                     mode, *args, **kwargs):
    """ Run the optimiser. """
    return cp_ga_optimiser.cp_ga_optimiser_from_raw_args(prob_funcs[0],
             domain_config_file, worker_manager, max_capital, mode=mode, *args, **kwargs)


class CPGAOPtimiserTestCase(CPGAOPtimiserTestCaseDefinitions,
                            CPOptimiserBaseTestCase,
                            BaseTestClass):
  """ Unit tests for GA optimiser on cartesian product spaces. """
  pass


if __name__ == '__main__':
  execute_tests()




"""
  Implements classes and methods to define domains for neural networks.
  The NNConstraintChecker class and its subclasses are used to check if a neural network
  architecture satisfies certain constriants. This is mostly needed to constrain the
  search space in NASBOT.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=too-many-arguments

import numpy as np
from copy import copy
# Local
from ..exd.domains import Domain


class NNConstraintChecker(object):
  """ A class for checking if a neural network satisfies constraints. """

  def __init__(self, max_num_layers, min_num_layers, max_mass, min_mass,
               max_in_degree, max_out_degree, max_num_edges,
               max_num_units_per_layer, min_num_units_per_layer):
    """ Constructor. """
    super(NNConstraintChecker, self).__init__()
    self.max_num_layers = max_num_layers
    self.min_num_layers = min_num_layers
    self.max_mass = max_mass
    self.min_mass = min_mass
    self.max_in_degree = max_in_degree
    self.max_out_degree = max_out_degree
    self.max_num_edges = max_num_edges
    self.max_num_units_per_layer = max_num_units_per_layer
    self.min_num_units_per_layer = min_num_units_per_layer
    self.constraint_names = ['max_num_layers', 'min_num_layers', 'max_mass',
      'min_mass', 'max_in_degree', 'max_out_degree', 'max_num_edges',
      'max_num_units_per_layer', 'min_num_units_per_layer']

  def __call__(self, nn, *args, **kwargs):
    """ Checks if the constraints are satisfied for the given nn. """
    return self.constraints_are_satisfied(nn, *args, **kwargs)

  def constraints_are_satisfied(self, nn, return_violation=False):
    """ Checks if the neural network nn satisfies the constraints. If return_violation
        is True, it returns a string representing the violation. """
    violation = ''
    if not self._check_leq_constraint(len(nn.layer_labels), self.max_num_layers):
      violation = 'too_many_layers'
    elif not self._check_geq_constraint(len(nn.layer_labels), self.min_num_layers):
      violation = 'too_few_layers'
    elif not self._check_leq_constraint(nn.get_total_mass(), self.max_mass):
      violation = 'too_much_mass'
    elif not self._check_geq_constraint(nn.get_total_mass(), self.min_mass):
      violation = 'too_little_mass'
    elif not self._check_leq_constraint(nn.get_out_degrees().max(), self.max_out_degree):
      violation = 'large_max_out_degree'
    elif not self._check_leq_constraint(nn.get_in_degrees().max(), self.max_in_degree):
      violation = 'large_max_in_degree'
    elif not self._check_leq_constraint(nn.conn_mat.sum(), self.max_num_edges):
      violation = 'too_many_edges'
    elif not self._check_leq_constraint(
                              self._finite_max_or_min(nn.num_units_in_each_layer, 1),
                              self.max_num_units_per_layer):
      violation = 'max_units_per_layer_exceeded'
    elif not self._check_geq_constraint(
                              self._finite_max_or_min(nn.num_units_in_each_layer, 0),
                              self.min_num_units_per_layer):
      violation = 'min_units_per_layer_not_exceeded'
    else:
      violation = self._child_constraints_are_satisfied(nn)
    return violation if return_violation else (violation == '')

  @classmethod
  def _check_leq_constraint(cls, value, bound):
    """ Returns true if bound is None or if value is less than or equal to bound. """
    return bound is None or (value <= bound)

  @classmethod
  def _check_geq_constraint(cls, value, bound):
    """ Returns true if bound is None or if value is greater than or equal to bound. """
    return bound is None or (value >= bound)

  def _child_constraints_are_satisfied(self, nn):
    """ Checks if the constraints of the child class are satisfied. """
    raise NotImplementedError('Implement in a child class.')

  @classmethod
  def _finite_max_or_min(cls, iterable, is_max):
    """ Returns the max ignorning Nones, nans and infs. """
    finite_vals = [x for x in iterable if x is not None and np.isfinite(x)]
    return max(finite_vals) if is_max else min(finite_vals)


class CNNConstraintChecker(NNConstraintChecker):
  """ A class for checking if a CNN satisfies constraints. """

  def __init__(self, max_num_layers, min_num_layers, max_mass, min_mass,
               max_in_degree, max_out_degree, max_num_edges,
               max_num_units_per_layer, min_num_units_per_layer,
               max_num_2strides=None):
    """ Constructor.
      max_num_2strides is the maximum number of 2-strides (either via pooling or conv
      operations) that the image can go through in the network.
    """
    super(CNNConstraintChecker, self).__init__(
      max_num_layers, min_num_layers, max_mass, min_mass,
      max_in_degree, max_out_degree, max_num_edges,
      max_num_units_per_layer, min_num_units_per_layer)
    self.max_num_2strides = max_num_2strides
    self.constraint_names.append('max_num_2strides')

  def _child_constraints_are_satisfied(self, nn):
    """ Checks if the constraints of the child class are satisfied. """
    img_inv_sizes = [piis for piis in nn.post_img_inv_sizes if piis != 'x']
    max_post_img_inv_sizes = None if self.max_num_2strides is None \
                                  else 2**self.max_num_2strides
    violation = ''
    if not self._check_leq_constraint(max(img_inv_sizes), max_post_img_inv_sizes):
      violation = 'too_many_2strides'
    return violation


class MLPConstraintChecker(NNConstraintChecker):
  """ A class for checking if a MLP satisfies constraints. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(MLPConstraintChecker, self).__init__(*args, **kwargs)

  def _child_constraints_are_satisfied(self, nn):
    """ Checks if the constraints of the child class are satisfied. """
    return ''


# An NN Domain class -----------------------------------------------------------------
class NNDomain(Domain):
  """ Domain for Neural Network architectures. """

  def __init__(self, nn_type, constraint_checker=None):
    """ Constructor. """
    self.nn_type = nn_type
    self.constraint_checker = constraint_checker
    super(NNDomain, self).__init__()

  def get_type(self):
    """ Returns type of the domain. """
    return "neural_network"

  def get_dim(self):
    """ Return dimension. """
    return 1

  def is_a_member(self, point):
    """ Returns true if point is in the domain. """
    if not self.nn_type == point.nn_class:
      return False
    else:
      return self.constraint_checker(point)

  @classmethod
  def members_are_equal(cls, point_1, point_2):
    """ Returns true if they are equal. """
    return neural_nets_are_equal(point_1, point_2)

  def __str__(self):
    """ Returns a string representation. """
    cc_attrs = {key:getattr(self.constraint_checker, key) for
                key in self.constraint_checker.constraint_names}
    return 'NN(%s):%s'%(self.nn_type, cc_attrs)


def neural_nets_are_equal(net1, net2):
  """ Returns true if both net1 and net2 are equal.
  """
  is_true = True
  for key in net1.__dict__.keys():
    val1 = net1.__dict__[key]
    val2 = net2.__dict__[key]
    is_true = True
    if isinstance(val1, dict):
      for val_key in val1.keys():
        is_true = is_true and np.all(val1[val_key] == val2[val_key])
    elif hasattr(val1, '__iter__'):
      is_true = is_true and np.all(val1 == val2)
    else:
      is_true = is_true and val1 == val2
    if not is_true: # break here if necessary
      return is_true
  return is_true


# An API to return an NN Domain using the constraints --------------------------------
def get_nn_domain_from_constraints(nn_type, constraint_dict):
  """ nn_type is the type of the network.
      See CNNConstraintChecker, MLPConstraintChecker, NNConstraintChecker constructors
      for args and kwargs.
  """
  constraint_dict = copy(constraint_dict)
  # Specify the mandatory and optional key values
  mandatory_keys = ['max_num_layers', 'max_mass']
  optional_key_vals = [('min_num_layers', 5),
                       ('min_mass', 0),
                       ('max_out_degree', np.inf),
                       ('max_in_degree', np.inf),
                       ('max_num_edges', np.inf),
                       ('max_num_units_per_layer', 10001),
                       ('min_num_units_per_layer', 5),
                      ]
  if nn_type.startswith('cnn'):
    optional_key_vals += [('max_num_2strides', np.inf)]
  # Check if the mandatory keys exist in constraint_dict
  for mkey in mandatory_keys:
    if mkey not in constraint_dict.keys():
      raise ValueError('Must specify keys %s in constraint_dict.'%(
                       ', '.join(mandatory_keys)))
  # If an optional key does not exist, then add it
  for okey, oval in optional_key_vals:
    if okey not in constraint_dict.keys():
      constraint_dict[okey] = oval
  # Specify the constructor
  if nn_type.startswith('cnn'):
    cc_constructor = CNNConstraintChecker
  elif nn_type.startswith('mlp'):
    cc_constructor = MLPConstraintChecker
  else:
    raise ValueError('Unknown nn_type: %s.'%(nn_type))
  # Now create constraint checker object
  cc_attributes = mandatory_keys + [okv[0] for okv in optional_key_vals]
  constraint_dict_to_pass = {key: constraint_dict[key] for key in cc_attributes}
  constraint_checker = cc_constructor(**constraint_dict_to_pass)
  # Create Domain and return
  return NNDomain(nn_type, constraint_checker)




"""
  Unit tests for EuclideanGPBandit.
  -- kandasamy@cs.cmu.edu
"""
import unittest

# Local imports
from ..gp.euclidean_gp import euclidean_gp_args
from . import gp_bandit
from .unittest_euclidean_random_optimiser import EuclideanOptimisersBaseTestCase, \
                                                    MFEuclideanOptimisersBaseTestCase
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.option_handler import load_options


@unittest.skip
class EuclideanGPBanditTestCase(EuclideanOptimisersBaseTestCase, BaseTestClass):
# class EuclideanGPBanditTestCase(EuclideanOptimisersBaseTestCase):
  """ Unit tests for the GP Bandit in Euclidean spaces. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager,
                                   options=None, reporter=None):
    """ Instantiates a GP bandit in Euclidean space. """
    return gp_bandit.EuclideanGPBandit(func_caller, worker_manager, is_mf=False,
             options=options, reporter=reporter)

  @classmethod
  def run_optimiser(cls, func_caller, worker_manager, max_capital, mode, *args, **kwargs):
    """ Runs optimiser. """
    return gp_bandit.gpb_from_func_caller(func_caller, worker_manager, max_capital,
             mode=mode, is_mf=False, *args, **kwargs)


# class EuclideanAddGPBanditTestCase(EuclideanOptimisersBaseTestCase, BaseTestClass):
class EuclideanAddGPBanditTestCase(EuclideanOptimisersBaseTestCase):
  """ Unit tests for Additive GP Bandit in Euclidean spaces. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager,
                                   options=None, reporter=None):
    """ Instantiates an additive GP bandit in Euclidean space. """
    if options is None:
      opt_args = gp_bandit.get_all_gp_bandit_args(euclidean_gp_args)
      options = load_options(opt_args)
    options.use_additive_gp = True
    return gp_bandit.EuclideanGPBandit(func_caller, worker_manager, is_mf=False,
             options=options, reporter=reporter)

  @classmethod
  def run_optimiser(cls, func_caller, worker_manager, max_capital, mode, *args, **kwargs):
    """ Runs optimiser. """
    return gp_bandit.gpb_from_func_caller(func_caller, worker_manager, max_capital,
             mode=mode, acq='add_ucb', is_mf=False, domain_add_max_group_size=0,
             *args, **kwargs)


@unittest.skip
class MFEuclideanGPBanditTestCase(MFEuclideanOptimisersBaseTestCase, BaseTestClass):
# class MFEuclideanGPBanditTestCase(MFEuclideanOptimisersBaseTestCase):
  """ Unit tests for multi-fidelity GP Bandit Optimiser in Euclidean spaces. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager,
                                   options=None, reporter=None):
    """ Instantiates a MFGP bandit in Euclidean space. """
    return gp_bandit.EuclideanGPBandit(func_caller, worker_manager, is_mf=True,
             options=options, reporter=reporter)

  @classmethod
  def run_optimiser(cls, func_caller, worker_manager, max_capital, mode, *args, **kwargs):
    """ Runs optimiser. """
    return gp_bandit.gpb_from_func_caller(func_caller, worker_manager, max_capital,
             mode=mode, is_mf=True, *args, **kwargs)


# class MFEuclideanAddGPBanditTestCase(MFEuclideanOptimisersBaseTestCase, BaseTestClass):
class MFEuclideanAddGPBanditTestCase(MFEuclideanOptimisersBaseTestCase):
  """ Unit tests for Additive GP Bandit in MF Euclidean spaces. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager,
                                   options=None, reporter=None):
    """ Instantiates an additive MF GP bandit in Euclidean space. """
    if options is None:
      opt_args = gp_bandit.get_all_gp_bandit_args(euclidean_gp_args)
      options = load_options(opt_args)
    options.use_additive_gp = True
    return gp_bandit.EuclideanGPBandit(func_caller, worker_manager, is_mf=False,
             options=options, reporter=reporter)

  @classmethod
  def run_optimiser(cls, func_caller, worker_manager, max_capital, mode, *args, **kwargs):
    """ Runs optimiser. """
    return gp_bandit.gpb_from_func_caller(func_caller, worker_manager, max_capital,
             mode=mode, acq='add_ucb', is_mf=True, domain_add_max_group_size=0,
             *args, **kwargs)


# class MFEuclideanESPGPBanditTestCase(MFEuclideanOptimisersBaseTestCase, BaseTestClass):
class MFEuclideanESPGPBanditTestCase(MFEuclideanOptimisersBaseTestCase):
  """ Unit tests for GP Bandit with ESP kernel in MF Euclidean spaces. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager,
                                   options=None, reporter=None):
    """ Instantiates an additive MF GP bandit in Euclidean space. """
    if options is None:
      opt_args = gp_bandit.get_all_gp_bandit_args(euclidean_gp_args)
      options = load_options(opt_args)
    options.domain_kernel_type = 'esp'
    return gp_bandit.EuclideanGPBandit(func_caller, worker_manager, is_mf=False,
             options=options, reporter=reporter)

  @classmethod
  def run_optimiser(cls, func_caller, worker_manager, max_capital, mode, *args, **kwargs):
    """ Runs optimiser. """
    return gp_bandit.gpb_from_func_caller(func_caller, worker_manager, max_capital,
             mode=mode, acq='add_ucb', is_mf=True, domain_add_max_group_size=0,
             *args, **kwargs)


if __name__ == '__main__':
  execute_tests()

